#    Copyrigh 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module for writing offset tables to file."""

from abc import ABCMeta, abstractmethod
from typing import Sequence

from pydicom.tag import ItemTag

from wsidicom.errors import WsiDicomBotOverflow
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.tags import ExtendedOffsetTableLengthsTag, ExtendedOffsetTableTag


class OffsetTableWriter(metaclass=ABCMeta):
    """Abstract class for writing offset tables to file."""

    def __init__(self, file: WsiDicomIO) -> None:
        """Initiate offset table writer.

        Parameters
        ----------
        file: WsiDicomIO
            File to write to.

        """
        self._file = file

    @abstractmethod
    def reserve(self, number_of_frames: int):
        """Reserve space in file for offset table.

        Parameters
        ----------
        number_of_frames: int
            Number of frames to reserve space for.

        """
        raise NotImplementedError()

    @abstractmethod
    def write(
        self,
        pixel_data_start: int,
        frame_positions: Sequence[int],
        last_frame_end: int,
    ) -> None:
        """Write offset table to file.

        Parameters
        ----------
        pixel_data_start: int
            File position of pixel data start
        frame_positions: Sequence[int]
            List of file positions for frames, relative to file start
        last_frame_end: int
            Position of last frame end.

        """
        raise NotImplementedError()


class BotWriter(OffsetTableWriter):
    def reserve(self, number_of_frames: int):
        self._table_start = self._file.tell()
        BYTES_PER_ITEM = 4
        tag_lengths = BYTES_PER_ITEM * number_of_frames
        self._file.write_tag(ItemTag)
        self._file.write_UL(tag_lengths)
        for _ in range(number_of_frames):
            self._file.write_UL(0)

    def write(
        self,
        pixel_data_start: int,
        frame_positions: Sequence[int],
        last_frame_end: int,
    ) -> None:
        BYTES_PER_ITEM = 4
        # Check that last BOT entry is not over 2^32 - 1
        last_entry = frame_positions[-1] - pixel_data_start
        if last_entry > 2**32 - 1:
            raise WsiDicomBotOverflow(
                "Image data exceeds 2^32 - 1 bytes "
                "An extended offset table should be used"
            )

        self._file.seek(self._table_start)  # Go to first BOT entry
        self._file.check_tag_and_length(
            ItemTag, BYTES_PER_ITEM * len(frame_positions), False, True
        )

        for frame_position in frame_positions:  # Write BOT
            self._file.write_UL(frame_position - pixel_data_start)


class EotWriter(OffsetTableWriter):
    def reserve(self, number_of_frames: int):
        self._table_start = self._file.tell()
        BYTES_PER_ITEM = 8
        eot_length = BYTES_PER_ITEM * number_of_frames
        self._file.write_tag_of_vr_and_length(ExtendedOffsetTableTag, "OV", eot_length)
        for _ in range(number_of_frames):
            self._file.write_unsigned_long_long(0)
        self._file.write_tag_of_vr_and_length(
            ExtendedOffsetTableLengthsTag, "OV", eot_length
        )
        for _ in range(number_of_frames):
            self._file.write_unsigned_long_long(0)

    def write(
        self,
        pixel_data_start: int,
        frame_positions: Sequence[int],
        last_frame_end: int,
    ) -> None:
        BYTES_PER_ITEM = 8
        # Check that last EOT entry is not over 2^64 - 1
        last_entry = frame_positions[-1] - pixel_data_start
        if last_entry > 2**64 - 1:
            raise ValueError(
                "Image data exceeds 2^64 - 1 bytes, likely something is wrong."
            )
        self._file.seek(self._table_start)  # Go to EOT table
        self._file.check_tag_and_length(
            ExtendedOffsetTableTag, BYTES_PER_ITEM * len(frame_positions), True, True
        )
        for frame_position in frame_positions:  # Write EOT
            relative_position = frame_position - pixel_data_start
            self._file.write_unsigned_long_long(relative_position)

        # EOT LENGTHS
        self._file.check_tag_and_length(
            ExtendedOffsetTableLengthsTag,
            BYTES_PER_ITEM * len(frame_positions),
            True,
            True,
        )
        frame_start = frame_positions[0]
        for frame_end in frame_positions[1:]:  # Write EOT lengths
            frame_length = frame_end - frame_start
            self._file.write_unsigned_long_long(frame_length)
            frame_start = frame_end

        # Last frame length, end does not include tag and length
        TAG_BYTES = 4
        LENGTH_BYTES = 4
        last_frame_start = frame_start + TAG_BYTES + LENGTH_BYTES
        last_frame_length = last_frame_end - last_frame_start
        self._file.write_unsigned_long_long(last_frame_length)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Sequence, Set, Tuple

from pydicom.uid import (
    JPEG2000,
    UID,
    ExplicitVRLittleEndian,
    JPEG2000Lossless,
    JPEGBaseline8Bit,
    JPEGLossless,
    JPEGLosslessSV1,
    JPEGLSLossless,
    JPEGLSNearLossless,
    RLELossless,
)

from wsidicom.codec import Codec
from wsidicom.config import settings
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.graphical_annotations import AnnotationInstance
from wsidicom.instance import ImageType, WsiDataset, WsiInstance
from wsidicom.source import Source
from wsidicom.thread import ConditionalThreadPoolExecutor
from wsidicom.web.wsidicom_web_client import WsiDicomWebClient
from wsidicom.web.wsidicom_web_image_data import WsiDicomWebImageData

"""A source for reading WSI DICOM files from DICOMWeb."""

# Transfer syntaxes to try in order of preference.
PREFERED_WEB_TRANSFER_SYNTAXES = [
    JPEGBaseline8Bit,
    JPEG2000,
    JPEG2000Lossless,
    JPEGLSNearLossless,
    JPEGLSLossless,
    JPEGLosslessSV1,
    JPEGLossless,
    RLELossless,
    ExplicitVRLittleEndian,
]


class WsiDicomWebSource(Source):
    """Source reading WSI DICOM instances from DICOMWeb."""

    def __init__(
        self,
        client: WsiDicomWebClient,
        study_uid: UID,
        series_uids: Iterable[UID],
        requested_transfer_syntaxes: Optional[Sequence[UID]] = None,
    ):
        """Create a WsiDicomWebSource.

        Parameters
        ----------
        client: WsiDicomWebClient
            Client use for DICOMWeb communication.
        study_uid: UID
            Study UID of DICOM WSI to open.
        series_uids: Union[str, UID, Iterable[Union[str, UID]]]
            Series UIDs of DICOM WSI top open.
        requested_transfer_syntaxes: Optional[Sequence[UID]] = None
            Transfer syntax to request for image data, for example
            UID("1.2.840.10008.1.2.4.50") for JPEGBaseline8Bit.

        """
        super().__init__()
        self._level_instances: List[WsiInstance] = []
        self._label_instances: List[WsiInstance] = []
        self._overview_instances: List[WsiInstance] = []
        self._annotation_instances: List[AnnotationInstance] = []
        detected_transfer_syntaxes_by_image_type: Dict[ImageType, Set[UID]] = (
            defaultdict(set)
        )

        def create_instance(
            uids: Tuple[UID, UID, UID, Optional[Set[UID]]]
        ) -> Optional[WsiInstance]:
            study_uid, series_uid, instance_uid, available_transfer_syntaxes = uids
            dataset = client.get_instance(study_uid, series_uid, instance_uid)
            if not WsiDataset.is_supported_wsi_dicom(dataset):
                logging.info(f"Non-supported instance {instance_uid}.")
                return
            dataset = WsiDataset(dataset)

            transfer_syntax = self._determine_transfer_syntax(
                client,
                dataset,
                available_transfer_syntaxes,
                detected_transfer_syntaxes_by_image_type[dataset.image_type],
                requested_transfer_syntaxes,
            )
            if transfer_syntax is None:
                logging.info(
                    f"No supported transfer syntax found for instance {uids[2]}."
                )
                return
            detected_transfer_syntaxes_by_image_type[dataset.image_type].add(
                transfer_syntax
            )
            image_data = WsiDicomWebImageData(
                client,
                dataset,
                transfer_syntax,
                self._decoded_frame_cache,
                self._encoded_frame_cache,
            )
            return WsiInstance(dataset, image_data)

        instance_uids = (
            (study_uid, series_uid, instance_uid, available_transfer_syntaxes)
            for series_uid, instance_uid, available_transfer_syntaxes in client.get_wsi_instances(
                study_uid, series_uids
            )
        )
        annotation_instances = (
            client.get_instance(study_uid, series_uid, instance_uid)
            for series_uid, instance_uid, _ in client.get_annotation_instances(
                study_uid, series_uids
            )
        )

        with ConditionalThreadPoolExecutor(settings.open_web_theads) as pool:
            instances = pool.map(create_instance, instance_uids)
            for instance in instances:
                if instance is None:
                    continue
                if instance.image_type == ImageType.VOLUME:
                    self._level_instances.append(instance)
                elif instance.image_type == ImageType.LABEL:
                    self._label_instances.append(instance)
                elif instance.image_type == ImageType.OVERVIEW:
                    self._overview_instances.append(instance)
            for annotation_instance in annotation_instances:
                self._annotation_instances.append(
                    AnnotationInstance.open_dataset(annotation_instance)
                )

        try:
            self._base_dataset = next(
                instance.dataset
                for instance in sorted(
                    self._level_instances,
                    reverse=True,
                    key=lambda instance: instance.dataset.image_size.width,
                )
            )

        except StopIteration:
            WsiDicomNotFoundError(
                "No level instances found", f"{study_uid}, {series_uids}"
            )
        self._base_tile_size = self._base_dataset.tile_size

    @property
    def base_dataset(self) -> WsiDataset:
        """The dataset of the one of the level instances."""
        return self._base_dataset

    @property
    def level_instances(self) -> List[WsiInstance]:
        """The level instances parsed from the source."""
        if settings.strict_tile_size_check:
            return [
                level
                for level in self._level_instances
                if level.dataset.tile_size == self._base_tile_size
            ]
        return self._level_instances

    @property
    def label_instances(self) -> List[WsiInstance]:
        """The label instances parsed from the source."""
        return self._label_instances

    @property
    def overview_instances(self) -> List[WsiInstance]:
        """The overview instances parsed from the source."""
        return self._overview_instances

    @property
    def annotation_instances(self) -> List[AnnotationInstance]:
        """The annotation instances parsed from the source."""
        return self._annotation_instances

    def close(self) -> None:
        pass

    def _determine_transfer_syntax(
        self,
        client: WsiDicomWebClient,
        dataset: WsiDataset,
        available_transfer_syntaxes: Optional[Set[UID]],
        detected_transfer_syntaxes: Set[UID],
        requested_transfer_syntaxes: Optional[Iterable[UID]],
    ) -> Optional[UID]:
        """Determine transfer syntax to use for image data.

        Parameters
        ----------
        client: WsiDicomWebClient
            Client used for DICOMWeb communication.
        dataset: WsiDataset
            Dataset to determine transfer syntax for.
        available_transfer_syntaxes: Optional[Set[UID]]
            Transfer syntaxes available on the server, or None if not known.
        detected_transfer_syntaxes: Set[UID]
            Transfer syntaxes that have already been detected.
        requested_transfer_syntaxes: Optional[Iterable[UID]]
            Transfer syntaxes to try in order of preference.

        Returns
        -------
        Optional[UID]
            Transfer syntax to use for image data, or None if no supported transfer
            syntax was found.
        """
        if requested_transfer_syntaxes is None:
            requested_transfer_syntaxes = PREFERED_WEB_TRANSFER_SYNTAXES

        supported_transfer_syntaxes = (
            transfer_syntax
            for transfer_syntax in requested_transfer_syntaxes
            if Codec.is_supported(
                transfer_syntax,
                dataset.samples_per_pixel,
                dataset.bits,
                dataset.photometric_interpretation,
            )
        )
        if available_transfer_syntaxes is not None:
            # Server provided list of available transfer syntaxes.
            # Select first one that is supported in order by user or default preference.
            transfer_syntax = next(
                (
                    transfer_syntax
                    for transfer_syntax in supported_transfer_syntaxes
                    if transfer_syntax in available_transfer_syntaxes
                ),
                None,
            )
            if transfer_syntax is not None:
                return transfer_syntax

        # No server provided list of available transfer syntaxes.
        # Detect if any of the supported transfer syntaxes are supported by the server.
        # Start with the ones that have already been detected.
        sorted_transfer_syntaxes = sorted(
            supported_transfer_syntaxes,
            key=lambda transfer_syntax: transfer_syntax in detected_transfer_syntaxes,
            reverse=True,
        )

        return next(
            (
                transfer_syntax
                for transfer_syntax in sorted_transfer_syntaxes
                if client.is_transfer_syntax_supported(
                    dataset.uids.slide.study_instance,
                    dataset.uids.slide.series_instance,
                    dataset.uids.instance,
                    transfer_syntax,
                )
            ),
            None,
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Json schema for Study model."""

from marshmallow import fields

from wsidicom.metadata.schema.common import LoadingSchema
from wsidicom.metadata.schema.json.fields import UidJsonField
from wsidicom.metadata.study import Study


class StudyJsonSchema(LoadingSchema[Study]):
    uid = UidJsonField(allow_none=True)
    identifier = fields.String(allow_none=True)
    date = fields.Date(allow_none=True)
    time = fields.Time(allow_none=True)
    accession_number = fields.String(allow_none=True)
    referring_physician_name = fields.String(allow_none=True)

    @property
    def load_type(self):
        return Study



#    Copyright 2022 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from typing import Optional

from PIL import Image as Pillow


class Settings:
    """Class containing settings. Settings are to be accessed through the
    global variable settings."""

    def __init__(self) -> None:
        self._strict_uid_check = False
        self._strict_tile_size_check = True
        self._strict_attribute_check = False
        self._strict_specimen_identifier_check = True
        self._focal_plane_distance_threshold = 0.000001
        self._pyramids_origin_threshold = 0.02
        self._prefered_decoder: Optional[str] = None
        self._open_web_theads: Optional[int] = None
        self._pillow_resampling_filter = Pillow.Resampling.BILINEAR
        self._ignore_specimen_preparation_step_on_validation_error = True
        self._truncate_long_dicom_strings = False
        self._decoded_frame_cache_size = 1000
        self._encoded_frame_cache_size = 1000
        self._level_scale_tolerance = 1e-2

    @property
    def strict_uid_check(self) -> bool:
        """If frame of reference uid needs to match."""
        return self._strict_uid_check

    @strict_uid_check.setter
    def strict_uid_check(self, value: bool) -> None:
        self._strict_uid_check = value

    @property
    def strict_attribute_check(self) -> bool:
        """If attribute marked with Requirement.STRICT is required."""
        return self._strict_attribute_check

    @strict_attribute_check.setter
    def strict_attribute_check(self, value: bool) -> None:
        self._strict_attribute_check = value

    @property
    def strict_specimen_identifier_check(self) -> bool:
        """If `True` the issuer of two specimen identifiers needs to match or both be
        None for the identifiers to match. If `False` the identifiers will match also if
        either issuer is None. Either way the identifier needs to match."""
        return self._strict_specimen_identifier_check

    @strict_specimen_identifier_check.setter
    def strict_specimen_identifier_check(self, value: bool) -> None:
        self._strict_specimen_identifier_check = value

    @property
    def strict_tile_size_check(self) -> bool:
        """If tile size need to match for levels. If `False` the tile size accross
        levels are allowed to be non-uniform."""
        return self._strict_tile_size_check

    @strict_tile_size_check.setter
    def strict_tile_size_check(self, value: bool) -> None:
        self._strict_tile_size_check = value

    @property
    def focal_plane_distance_threshold(self) -> float:
        """Threshold in mm for which distances between focal planes are
        considered to be equal. Default is 1 nm, as distance between focal
        planes are likely larger than this.
        """
        return self._focal_plane_distance_threshold

    @focal_plane_distance_threshold.setter
    def focal_plane_distance_threshold(self, value: float) -> None:
        self._focal_plane_distance_threshold = value

    @property
    def pyramids_origin_threshold(self) -> float:
        """Threshold in mm for the distance between origins of instances
        to group them into the same pyramid. Default is 0.02 mm.
        """
        return self._pyramids_origin_threshold

    @pyramids_origin_threshold.setter
    def pyramids_origin_threshold(self, value: float) -> None:
        self._pyramids_origin_threshold = value

    @property
    def prefered_decoder(self) -> Optional[str]:
        """Name of preferred decoder to use."""
        return self._prefered_decoder

    @prefered_decoder.setter
    def prefered_decoder(self, value: Optional[str]) -> None:
        self._prefered_decoder = value

    @property
    def open_web_theads(self) -> Optional[int]:
        """Number of threads to use when opening web instances."""
        return self._open_web_theads

    @open_web_theads.setter
    def open_web_theads(self, value: Optional[int]) -> None:
        self._open_web_theads = value

    @property
    def pillow_resampling_filter(self) -> Pillow.Resampling:
        """The resampling filter to use when rescaling images."""
        return self._pillow_resampling_filter

    @pillow_resampling_filter.setter
    def pillow_resampling_filter(self, value: Pillow.Resampling) -> None:
        self._pillow_resampling_filter = value

    @property
    def ignore_specimen_preparation_step_on_validation_error(self) -> bool:
        """If ignore specimen preparation steps that fails to validate. If false all
        steps will be ignored if one fails to validate."""
        return self._ignore_specimen_preparation_step_on_validation_error

    @ignore_specimen_preparation_step_on_validation_error.setter
    def ignore_specimen_preparation_step_on_validation_error(self, value: bool) -> None:
        self._ignore_specimen_preparation_step_on_validation_error = value

    @property
    def truncate_long_dicom_strings_on_validation_error(self) -> bool:
        """If long DICOM strings should be truncated. This is only used if
        `pydicom.settings.writing_validation_mode` is set to `pydicom.config.RAISE`. If
        set to `True` long strings will be truncated if needed to pass validation."""
        return self._truncate_long_dicom_strings

    @truncate_long_dicom_strings_on_validation_error.setter
    def truncate_long_dicom_strings_on_validation_error(self, value: bool) -> None:
        self._truncate_long_dicom_strings = value

    @property
    def decoded_frame_cache_size(self) -> int:
        """Size of the decoded frame cache."""
        return self._decoded_frame_cache_size

    @decoded_frame_cache_size.setter
    def decoded_frame_cache_size(self, value: int) -> None:
        self._decoded_frame_cache_size = value

    @property
    def encoded_frame_cache_size(self) -> int:
        """Size of the encoded frame cache."""
        return self._encoded_frame_cache_size

    @encoded_frame_cache_size.setter
    def encoded_frame_cache_size(self, value: int) -> None:
        self._encoded_frame_cache_size = value

    @property
    def level_scale_tolerance(self) -> float:
        """Tolerance for level scale comparison. Default is 1e-2."""
        return self._level_scale_tolerance

    @level_scale_tolerance.setter
    def level_scale_tolerance(self, value: float) -> None:
        self._level_scale_tolerance = value


settings = Settings()
"""Global settings variable."""



#    Copyright 2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from collections.abc import Iterator
from threading import Lock
from typing import Callable, Dict, Generic, Iterable, Optional, Sequence, Tuple, TypeVar

from PIL.Image import Image

CacheKeyType = TypeVar("CacheKeyType")
CacheItemType = TypeVar("CacheItemType")


class LRU(Generic[CacheKeyType, CacheItemType]):
    def __init__(self, maxsize: int):
        self._lock = Lock()
        self._cache: Dict[CacheKeyType, CacheItemType] = {}
        self._maxsize = maxsize

    @property
    def maxsize(self) -> int:
        return self._maxsize

    def get(self, key: CacheKeyType) -> Optional[CacheItemType]:
        with self._lock:
            item = self._cache.pop(key, None)
            if item is not None:
                self._cache[key] = item
            return item

    def put(self, key: CacheKeyType, value: CacheItemType) -> None:
        with self._lock:
            self._cache[key] = value
            if len(self._cache) > self._maxsize:
                self._cache.pop(next(iter(self._cache)))

    def clear(self) -> None:
        with self._lock:
            self._cache.clear()

    def resize(self, maxsize: int) -> None:
        with self._lock:
            self._maxsize = maxsize
            if len(self._cache) > maxsize:
                for _ in range(len(self._cache) - maxsize):
                    self._cache.pop(next(iter(self._cache)))


class FrameCache(Generic[CacheItemType]):
    def __init__(self, size: int):
        self._size = size
        self._lru_cache = LRU[Tuple[int, int], CacheItemType](size)

    def get_tile_frame(
        self,
        image_data_id: int,
        frame_index: int,
        frame_getter: Callable[[int], CacheItemType],
    ) -> CacheItemType:
        if self._lru_cache.maxsize < 1:
            return frame_getter(frame_index)
        frame = self._lru_cache.get((image_data_id, frame_index))
        if frame is None:
            frame = frame_getter(frame_index)
            self._lru_cache.put((image_data_id, frame_index), frame)
        return frame

    def get_tile_frames(
        self,
        image_data_id: int,
        frame_indices: Sequence[int],
        frames_getter: Callable[[Iterable[int]], Iterator[CacheItemType]],
    ) -> Iterator[CacheItemType]:
        if self._lru_cache.maxsize < 1:
            return frames_getter(frame_indices)
        cached_frames = {
            frame_index: frame
            for (frame_index, frame) in (
                (frame_index, self._lru_cache.get((image_data_id, frame_index)))
                for frame_index in frame_indices
            )
            if frame is not None
        }
        fetched_frames = frames_getter(
            frame_index
            for frame_index in frame_indices
            if frame_index not in cached_frames
        )
        for frame_index in frame_indices:
            frame = cached_frames.get(frame_index)
            if frame is None:
                frame = next(fetched_frames)
                self._lru_cache.put((image_data_id, frame_index), frame)
            yield frame

    def clear(self) -> None:
        self._lru_cache.clear()

    def resize(self, size: int) -> None:
        self._lru_cache.resize(size)


EncodedFrameCache = FrameCache[bytes]
DecodedFrameCache = FrameCache[Image]



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math
from typing import Iterable, Optional

from PIL.Image import Image

from wsidicom.config import settings
from wsidicom.errors import WsiDicomNoResolutionError, WsiDicomOutOfBoundsError
from wsidicom.geometry import Point, Region, Size, SizeMm
from wsidicom.group.group import Group
from wsidicom.instance import WsiInstance
from wsidicom.stringprinting import dict_pretty_str


class Level(Group):
    """Represents a level in the pyramid and contains one or more instances
    having the same pyramid level index, pixel spacing, and size but possibly
    different focal planes and/or optical paths.
    """

    def __init__(self, instances: Iterable[WsiInstance], base_pixel_spacing: SizeMm):
        """Create a level from list of WsiInstances. Assign the pyramid level
        index from pixel spacing of base level.

        Parameters
        ----------
        instances: Sequence[WsiInstance]
            Instances to build the level.
        base_pixel_spacing: SizeMm
            Pixel spacing of base level.
        """
        super().__init__(instances)
        self._base_pixel_spacing = base_pixel_spacing
        self._level = self._assign_level(self._base_pixel_spacing)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.instances}, " f"{self._base_pixel_spacing})"
        )

    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        string = f"Level: {self.level}, size: {self.size} px, mpp: {self.mpp} um/px"
        if depth is not None:
            depth -= 1
            if depth < 0:
                return string
        string += " Instances: " + dict_pretty_str(self.instances, indent, depth)
        return string

    @property
    def pyramid(self) -> str:
        """Return string representation of the level"""
        return (
            f"Level [{self.level}]"
            f" tiles: {self.default_instance.tiled_size},"
            f" size: {self.size}, mpp: {self.mpp} um/px"
        )

    @property
    def tile_size(self) -> Size:
        return self.default_instance.tile_size

    @property
    def level(self) -> int:
        """Return pyramid level"""
        return self._level

    @property
    def mpp(self) -> SizeMm:
        if self.pixel_spacing is None:
            raise WsiDicomNoResolutionError()
        return self.pixel_spacing * 1000.0

    @property
    def pixel_spacing(self) -> SizeMm:
        if self._pixel_spacing is None:
            raise WsiDicomNoResolutionError()
        return self._pixel_spacing

    def matches(self, other_level: "Level") -> bool:
        """Check if level matches other level. If strict common Uids should
        match. Wsi type and tile size should always match.

        Parameters
        ----------
        other_level: Group
            Other level to match against.

        Returns
        -------
        bool
            True if other level matches.
        """
        return (
            self.uids.matches(other_level.uids)
            and other_level.image_type == self.image_type
            and (
                not settings.strict_tile_size_check
                or other_level.tile_size == self.tile_size
            )
        )

    def get_highest_level(self) -> int:
        """Return lowest level that produces a single pixel sized image.

        Returns
        -------
        int
            Relative level where the pixel size becomes 1x1.
        """
        return math.ceil(math.log2(max(self.size.width, self.size.height)))

    def get_lowest_single_tile_level(self) -> int:
        """Return lowest level that produces a single tile sized image.

        Returns
        -------
        int
            Relative level where the pixel size becomes a single tile size.
        """
        return math.ceil(
            math.log2(
                max(
                    self.size.width / self.tile_size.width,
                    self.size.height / self.tile_size.height,
                )
            )
        )

    def get_scaled_tile(
        self,
        tile: Point,
        level: int,
        z: Optional[float] = None,
        path: Optional[str] = None,
    ) -> Image:
        """Return tile in another level by scaling a region.
        If the tile is an edge tile, the resulting tile is croped
        to remove part outside of the image (as defined by level size).

        Parameters
        ----------
        tile: Point
            Non scaled tile coordinate
        level: int
            Level to scale from
        z: Optional[float] = None
            Z coordinate
        path: Optional[str] = None
            Optical path

        Returns
        -------
        Image
            A tile image
        """
        scale = self.calculate_scale(level)
        instance = self.get_instance(z, path)
        scaled_region = Region.from_tile(tile, instance.tile_size) * scale
        cropped_region = scaled_region.crop(instance.image_data.image_size)
        if not self.valid_pixels(cropped_region):
            raise WsiDicomOutOfBoundsError(
                f"Region {cropped_region}", f"level size {self.size}"
            )
        image = self.get_region(cropped_region, z, path)
        tile_size = cropped_region.size.ceil_div(scale)
        image = image.resize(
            tile_size.to_tuple(), resample=settings.pillow_resampling_filter
        )
        return image

    def get_scaled_encoded_tile(
        self,
        tile: Point,
        scale: int,
        z: Optional[float] = None,
        path: Optional[str] = None,
    ) -> bytes:
        """Return encoded tile in another level by scaling a region.

        Parameters
        ----------
        tile: Point
            Non scaled tile coordinate
        level: int
           Level to scale from
        z: Optional[float] = None
            Z coordinate
        path: Optional[str] = None
            Optical path

        Returns
        -------
        bytes
            A transfer syntax encoded tile
        """
        image = self.get_scaled_tile(tile, scale, z, path)
        instance = self.get_instance(z, path)
        return instance.image_data.encoder.encode(image)

    def calculate_scale(self, level_to: int) -> int:
        """Return scaling factor to given level.

        Parameters
        ----------
        level_to -- index of level to scale to

        Returns
        -------
        int
            Scaling factor between this level and given level
        """
        return int(2 ** (level_to - self.level))

    def _assign_level(self, base_pixel_spacing: SizeMm) -> int:
        """Return (2^level scale factor) based on pixel spacing.
        Will round to closest integer. Raises NotImplementedError if level is
        to far from integer.

        Parameters
        ----------
        base_pixel_spacing: SizeMm
            The pixel spacing of the base level

        Returns
        -------
        int
            The pyramid order of the level
        """
        float_level = math.log2(self.pixel_spacing.width / base_pixel_spacing.width)
        level = int(round(float_level))
        TOLERANCE = 1e-2
        if not math.isclose(float_level, level, abs_tol=settings.level_scale_tolerance):
            raise NotImplementedError(
                f"Levels needs to be integer. Got {float_level} that is more than set"
                f"tolerance {TOLERANCE} from the closest integer {level}. "
                f"Base spacing is {base_pixel_spacing}, this level has spacing "
                f"{self.pixel_spacing}.",
            )
        return level


class BaseLevel(Level):
    def __init__(self, instances: Iterable[WsiInstance]):
        """Create a base level from list of WsiInstances.

        Parameters
        ----------
        instances: Sequence[WsiInstance]
            Instances to build the level.
        """
        pixel_spacing = next(iter(instances)).pixel_spacing
        if pixel_spacing is None:
            raise WsiDicomNoResolutionError()
        super().__init__(instances, pixel_spacing)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing image codecs."""
from wsidicom.codec.codec import Codec
from wsidicom.codec.decoder import Decoder
from wsidicom.codec.encoder import Encoder, LossyCompressionIsoStandard
from wsidicom.codec.media_types import determine_media_type
from wsidicom.codec.settings import (
    Channels,
    Jpeg2kSettings,
    JpegLosslessSettings,
    JpegLsSettings,
    JpegSettings,
    NumpySettings,
    RleSettings,
    Settings,
    Subsampling,
)

__all__ = [
    "Codec",
    "Decoder",
    "Encoder",
    "LossyCompressionIsoStandard",
    "determine_media_type",
    "Channels",
    "Jpeg2kSettings",
    "JpegLosslessSettings",
    "JpegLsSettings",
    "JpegSettings",
    "NumpySettings",
    "RleSettings",
    "Settings",
    "Subsampling",
]



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Iterable

from wsidicom.cache import DecodedFrameCache, EncodedFrameCache
from wsidicom.config import settings
from wsidicom.graphical_annotations import AnnotationInstance
from wsidicom.instance import WsiDataset, WsiInstance

"""A Source provides image instances that can be opened by WsiDicom. Implement this
class to extend support to other DICOM sources or to read other WSI formats.
"""


class Source(metaclass=ABCMeta):
    """A source providing DICOM WSI instances to open.

    A source should be initiated with a path or similar, and parse the content into
    instances.
    """

    def __init__(self):
        self._decoded_frame_cache = DecodedFrameCache(settings.decoded_frame_cache_size)
        self._encoded_frame_cache = EncodedFrameCache(settings.encoded_frame_cache_size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    @abstractmethod
    def base_dataset(self) -> WsiDataset:
        """Return a representative dataset for the source content."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def level_instances(self) -> Iterable[WsiInstance]:
        """Return all level instances from the source."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def label_instances(self) -> Iterable[WsiInstance]:
        """Return all label instances from the source."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def overview_instances(self) -> Iterable[WsiInstance]:
        """Return all overview instances from the source."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def annotation_instances(self) -> Iterable[AnnotationInstance]:
        """Return all annotation instances from the source."""
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        """Close any opened resources (such as files)."""
        raise NotImplementedError()

    def clear_cache(self) -> None:
        """Clear the frame caches."""
        self._decoded_frame_cache.clear()
        self._encoded_frame_cache.clear()

    def resize_cache(self, size: int) -> None:
        """Clear the frame caches."""
        self._decoded_frame_cache.resize(size)
        self._encoded_frame_cache.resize(size)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Patient model."""

from collections import defaultdict
from typing import Any, Dict, Type

from marshmallow import fields, post_load, pre_dump
from pydicom.sr.coding import Code
from pydicom.valuerep import VR

from wsidicom.metadata.patient import Patient, PatientDeIdentification, PatientSex
from wsidicom.metadata.schema.dicom.fields import (
    BooleanDicomField,
    CodeDicomField,
    DateDicomField,
    DefaultingNoneDicomField,
    EnumDicomField,
    FlattenOnDumpNestedDicomField,
    ListDicomField,
    PersonNameDicomField,
    SingleCodeSequenceField,
    StringDicomField,
)
from wsidicom.metadata.schema.dicom.schema import (
    DicomSchema,
    ModuleDicomSchema,
)


class PatientDeIdentificationDicomSchema(DicomSchema[PatientDeIdentification]):
    identity_removed = BooleanDicomField(data_key="PatientIdentityRemoved")
    method_strings = ListDicomField(
        StringDicomField(VR.LO), data_key="DeidentificationMethod"
    )
    method_codes = fields.List(
        CodeDicomField(Code), data_key="DeidentificationMethodCodeSequence"
    )

    @property
    def load_type(self) -> Type[PatientDeIdentification]:
        return PatientDeIdentification

    @pre_dump
    def pre_dump(self, de_identification: PatientDeIdentification, **kwargs):
        fields = {"identity_removed": de_identification.identity_removed}
        if de_identification.methods is not None:
            de_identification_fields = defaultdict(list)
            for method in de_identification.methods:
                if isinstance(method, str):
                    de_identification_fields["method_strings"].append(method)
                else:
                    de_identification_fields["method_codes"].append(method)
            fields.update(de_identification_fields)
        return fields

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        method_strings = data.pop("method_strings", [])
        method_codes = data.pop("method_codes", [])
        methods = method_strings + method_codes
        if len(methods) > 0:
            data["methods"] = methods
        return super().post_load(data, **kwargs)


class PatientDicomSchema(ModuleDicomSchema[Patient]):
    name = DefaultingNoneDicomField(
        PersonNameDicomField(), data_key="PatientName", load_default=None
    )
    identifier = DefaultingNoneDicomField(
        StringDicomField(value_representation=VR.LO),
        data_key="PatientID",
        load_default=None,
    )
    birth_date = DefaultingNoneDicomField(
        DateDicomField(), data_key="PatientBirthDate", load_default=None
    )
    sex = DefaultingNoneDicomField(
        EnumDicomField(PatientSex), data_key="PatientSex", load_default=None
    )
    species_description_string = StringDicomField(
        value_representation=VR.LO,
        data_key="PatientSpeciesDescription",
        allow_none=True,
    )
    species_description_code = SingleCodeSequenceField(
        Code, data_key="PatientSpeciesCodeSequence", allow_none=True
    )
    de_identification = FlattenOnDumpNestedDicomField(
        PatientDeIdentificationDicomSchema(), allow_none=True
    )

    @property
    def load_type(self) -> Type[Patient]:
        return Patient

    @pre_dump
    def pre_dump(self, patient: Patient, **kwargs):
        fields = {
            "name": patient.name,
            "identifier": patient.identifier,
            "birth_date": patient.birth_date,
            "sex": patient.sex,
            "de_identification": patient.de_identification,
        }

        if isinstance(patient.species_description, str):
            fields["species_description_string"] = patient.species_description
        elif isinstance(patient.species_description, Code):
            fields["species_description_code"] = patient.species_description
        return fields

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        species_description_string = data.pop("species_description_string", None)
        species_description_code = data.pop("species_description_code", None)
        if species_description_code is not None:
            data["species_description"] = species_description_code
        elif species_description_string is not None:
            data["species_description"] = species_description_string
        return super().post_load(data, **kwargs)

    @property
    def module_name(self) -> str:
        return "patient"



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Iterable, Iterator

from PIL.Image import Image
from pydicom.uid import UID

from wsidicom.cache import DecodedFrameCache, EncodedFrameCache
from wsidicom.codec import Codec
from wsidicom.geometry import Point
from wsidicom.instance import WsiDataset, WsiDicomImageData
from wsidicom.web.wsidicom_web_client import WsiDicomWebClient


class WsiDicomWebImageData(WsiDicomImageData):
    """
    ImageData for WSI DICOM instances read from DICOMWeb.

    Overrides get_decoded_tiles() and get_encoded_tiles() to fetch multiple frames.
    """

    def __init__(
        self,
        client: WsiDicomWebClient,
        dataset: WsiDataset,
        transfer_syntax: UID,
        decoded_frame_cache: DecodedFrameCache,
        encoded_frame_cache: EncodedFrameCache,
    ):
        """
        Create WsiDicomWebImageData from  dataset and provided client.

        Parameters
        ----------
        client: WsiDicomWebClient
            DICOMWeb client for reading image data.
        dataset: WsiDataset
            Dataset for the image data.
        transfer_syntax: UID
            Transfer syntax to request for image data, for example
            UID("1.2.840.10008.1.2.4.50") for JPEGBaseline8Bit.
        """
        self._client = client
        self._study_uid = dataset.uids.slide.study_instance
        self._series_uid = dataset.uids.slide.series_instance
        self._instance_uid = dataset.uids.instance
        self._transfer_syntax = transfer_syntax
        codec = Codec.create(
            self.transfer_syntax,
            dataset.samples_per_pixel,
            dataset.bits,
            dataset.tile_size,
            dataset.photometric_interpretation,
        )
        super().__init__([dataset], codec, decoded_frame_cache, encoded_frame_cache)

    @property
    def transfer_syntax(self) -> UID:
        """The uid of the transfer syntax of the image."""
        return self._transfer_syntax

    def _get_decoded_tiles(
        self, tiles: Iterable[Point], z: float, path: str
    ) -> Iterator[Image]:
        """
        Return Pillow images for tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Iterator[Image]
            Tiles as Images.
        """
        frame_indices = [self._get_frame_index(tile, z, path) for tile in tiles]
        frames = self._decoded_frame_cache.get_tile_frames(
            id(self),
            [frame_index for frame_index in frame_indices if frame_index != -1],
            self._get_decoded_tile_frames,
        )

        for frame_index in frame_indices:
            if frame_index == -1:
                yield self.blank_tile
            else:
                yield next(frames)

    def _get_encoded_tiles(
        self, tiles: Iterable[Point], z: float, path: str
    ) -> Iterator[bytes]:
        """
        Return bytes for tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Iterator[Image]
            Tiles as Images.
        """
        frame_indices = [self._get_frame_index(tile, z, path) for tile in tiles]
        frames = self._encoded_frame_cache.get_tile_frames(
            id(self),
            [frame_index for frame_index in frame_indices if frame_index != -1],
            self._get_tile_frames,
        )
        for frame_index in frame_indices:
            if frame_index == -1:
                yield self.blank_encoded_tile
            else:
                yield next(frames)

    def _get_tile_frame(self, frame_index: int) -> bytes:
        # First frame for DICOM web is 1.
        return next(
            self._client.get_frames(
                self._study_uid,
                self._series_uid,
                self._instance_uid,
                [frame_index + 1],
                self._transfer_syntax,
            )
        )

    def _get_tile_frames(self, frame_indices: Iterable[int]) -> Iterator[bytes]:
        # First frame for DICOM web is 1.
        return self._client.get_frames(
            self._study_uid,
            self._series_uid,
            self._instance_uid,
            [frame_index + 1 for frame_index in frame_indices],
            self._transfer_syntax,
        )

    def _get_decoded_tile_frames(self, frame_indices: Iterable[int]) -> Iterator[Image]:
        for frame in self._get_tile_frames(frame_indices):
            yield self.decoder.decode(frame)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Json schema for Image model."""

from marshmallow import fields

from wsidicom.codec.encoder import LossyCompressionIsoStandard
from wsidicom.metadata.image import (
    ExtendedDepthOfField,
    FocusMethod,
    Image,
    ImageCoordinateSystem,
    LossyCompression,
)
from wsidicom.metadata.schema.common import LoadingSchema
from wsidicom.metadata.schema.json.fields import PointMmJsonField, SizeMmJsonField


class ExtendedDepthOfFieldJsonSchema(LoadingSchema[ExtendedDepthOfField]):
    number_of_focal_planes = fields.Integer()
    distance_between_focal_planes = fields.Float()

    @property
    def load_type(self):
        return ExtendedDepthOfField


class ImageCoordinateSystemJsonSchema(LoadingSchema[ImageCoordinateSystem]):
    origin = PointMmJsonField()
    rotation = fields.Float()
    z_offset = fields.Float(allow_none=True)

    @property
    def load_type(self):
        return ImageCoordinateSystem


class LossyCompressionJsonSchema(LoadingSchema[LossyCompression]):
    method = fields.Enum(LossyCompressionIsoStandard, by_value=True)
    ratio = fields.Float()

    @property
    def load_type(self):
        return LossyCompression


class ImageJsonSchema(LoadingSchema[Image]):
    acquisition_datetime = fields.DateTime(allow_none=True)
    focus_method = fields.Enum(FocusMethod, by_value=True, allow_none=True)
    extended_depth_of_field = fields.Nested(
        ExtendedDepthOfFieldJsonSchema(), allow_none=True
    )
    image_coordinate_system = fields.Nested(
        ImageCoordinateSystemJsonSchema(), allow_none=True
    )
    pixel_spacing = SizeMmJsonField(allow_none=True)
    focal_plane_spacing = fields.Float(allow_none=True)
    depth_of_field = fields.Float(allow_none=True)
    lossy_compressions = fields.List(
        fields.Nested(LossyCompressionJsonSchema()), allow_none=True
    )

    @property
    def load_type(self):
        return Image



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Abstract frame index for files with encapsulated data."""

from typing import Optional

from pydicom.tag import ItemTag

from wsidicom.errors import WsiDicomFileError
from wsidicom.file.io.frame_index.parser import FrameIndexParser


class EncapsulatedPixelDataFrameIndexParser(FrameIndexParser):
    def _validate_pixel_data_start(self):
        """Check that pixel data tag is present and that the tag length is
        set as undefined. Raises WsiDicomFileError otherwise.

        """
        super()._validate_pixel_data_start(None)

    def _read_bot_length(self) -> Optional[int]:
        """Read the length of the basic table offset (BOT). Returns None if BOT
        is empty.

        Returns
        -------
        Optional[int]
            BOT length.
        """
        BOT_BYTES = 4
        if self._file.read_tag() != ItemTag:
            raise WsiDicomFileError(
                str(self._file), "Basic offset table did not start with an ItemTag"
            )
        bot_length = self._file.read_UL()
        if bot_length == 0:
            return None
        elif bot_length % BOT_BYTES:
            raise WsiDicomFileError(
                str(self._file),
                f"Basic offset table should be a multiple of {BOT_BYTES} bytes",
            )
        return bot_length



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from PIL.Image import Image
from pydicom.uid import UID, generate_uid
from upath import UPath

from wsidicom.codec import Encoder
from wsidicom.codec import Settings as EncoderSettings
from wsidicom.config import settings
from wsidicom.errors import (
    WsiDicomMatchError,
    WsiDicomNotFoundError,
    WsiDicomOutOfBoundsError,
)
from wsidicom.file import OffsetTableType, WsiDicomFileSource, WsiDicomFileTarget
from wsidicom.geometry import Point, PointMm, Region, RegionMm, Size, SizeMm
from wsidicom.graphical_annotations import AnnotationInstance
from wsidicom.instance import WsiDataset, WsiInstance
from wsidicom.metadata.wsi import WsiMetadata
from wsidicom.series import Labels, Overviews, Pyramid, Pyramids
from wsidicom.source import Source
from wsidicom.stringprinting import list_pretty_str
from wsidicom.uid import SlideUids
from wsidicom.web import WsiDicomWebClient, WsiDicomWebSource


class WsiDicom:
    """A WSI containing pyramidal levels and optionally labels and/or overviews."""

    def __init__(
        self,
        source: Source,
        source_owned: bool = True,
    ):
        """Hold WSI DICOM levels, labels and overviews.

        Note that WsiDicom.open() should be used for opening DICOM WSI files.

        Parameters
        ----------
        source: Source
            A source providing instances for the wsi to open.
        source_owned: bool = True
            If source should be closed by this instance if used in a context manager.
        """
        self._selected_pyramid = 0
        self._source = source
        self._source_owned = source_owned
        self._pyramids = Pyramids.open(source.level_instances)
        self._labels = Labels.open(source.label_instances)
        self._overviews = Overviews.open(source.overview_instances)
        self._annotations = list(source.annotation_instances)
        self._uids = self._validate_collection()

    @classmethod
    def open(
        cls,
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]],
        file_options: Optional[Dict[str, Any]] = None,
    ) -> "WsiDicom":
        """Open valid WSI DICOM files and return a WsiDicom object.

        Non-valid files are ignored.

        Parameters
        ----------
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]]
            Files to open. Can be a path for a single file, a list of paths for multiple
            files, or a path to a folder containing files. Path can be local or an URL
            supported by fsspec.
        file_options: Optional[Dict[str, Any]] = None
            Optional options for when opening files.

        Returns
        -------
        WsiDicom
            WsiDicom created from WSI DICOM files in path.
        """
        source = WsiDicomFileSource.open(files, file_options)
        return cls(source, True)

    @classmethod
    def open_dicomdir(
        cls, path: UPath, file_options: Optional[Dict[str, Any]] = None
    ) -> "WsiDicom":
        """Open WSI DICOM files in DICOMDIR and return a WsiDicom object.

        Parameters
        ----------
        path: UPath
            Path to DICOMDIR file or directory with a DICOMDIR file. Path can be local
            or an URL supported by fsspec.
        file_options: Optional[Dict[str, Any]] = None
            Optional options for when opening files.

        Returns
        -------
        WsiDicom
            WsiDicom created from WSI DICOM files in DICOMDIR.
        """
        source = WsiDicomFileSource.open_dicomdir(path, file_options)
        return cls(source, True)

    @classmethod
    def open_streams(
        cls,
        streams: Iterable[BinaryIO],
    ) -> "WsiDicom":
        """Open valid WSI DICOM files in path or stream and return a WsiDicom object.

        Non-valid files are ignored. Only opened files (i.e. not streams) will e closed
        by WsiDicom.

        Parameters
        ----------
        streams: Iterable[BinaryIO],
            Streams to open.

        Returns
        -------
        WsiDicom
            WsiDicom created from WSI DICOM files in path.
        """
        source = WsiDicomFileSource.open_streams(streams)
        return cls(source, False)

    @classmethod
    def open_web(
        cls,
        client: WsiDicomWebClient,
        study_uid: Union[str, UID],
        series_uids: Union[str, UID, Iterable[Union[str, UID]]],
        requested_transfer_syntax: Optional[
            Union[str, UID, Sequence[Union[str, UID]]]
        ] = None,
    ) -> "WsiDicom":
        """Open WSI DICOM instances using DICOM web client.

        Parameters
        ----------
        client: WsiDicomWebClient
            Configured DICOM web client.
        study_uid: Union[str, UID]
            Study uid of wsi to open.
        series_uids: Union[str, UID, Iterable[Union[str, UID]]]
            Series uids of wsi to open
        requested_transfer_syntax: Optional[
            Union[str, UID, Sequence[Union[str, UID]]]
        ] = JPEGBaseline8Bit
            Transfer syntax to request for image data, for example
            "1.2.840.10008.1.2.4.50" for JPEGBaseline8Bit. By default the first
            supported transfer syntax is requested.

        Returns
        -------
        WsiDicom
            WsiDicom created from WSI DICOM instances in study-series.
        """
        study_uid = UID(study_uid)
        if isinstance(series_uids, (str, UID)):
            series_uids = [UID(series_uids)]
        else:
            series_uids = [UID(series_uid) for series_uid in series_uids]
        if isinstance(requested_transfer_syntax, (str, UID)):
            requested_transfer_syntax = [UID(requested_transfer_syntax)]
        elif requested_transfer_syntax is not None:
            requested_transfer_syntax = [
                UID(transfer_syntax) for transfer_syntax in requested_transfer_syntax
            ]

        source = WsiDicomWebSource(
            client, study_uid, series_uids, requested_transfer_syntax
        )
        return cls(source, True)

    def save(
        self,
        output_path: Union[str, Path, UPath],
        uid_generator: Callable[..., UID] = generate_uid,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        offset_table: Optional[Union["str", OffsetTableType]] = None,
        include_pyramids: Optional[Sequence[int]] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_labels: bool = True,
        include_overviews: bool = True,
        add_missing_levels: bool = False,
        label: Optional[Union[Image, Union[str, Path, UPath]]] = None,
        transcoding: Optional[Union[EncoderSettings, Encoder]] = None,
        file_options: Optional[Dict[str, Any]] = None,
    ) -> List[UPath]:
        """
        Save wsi as DICOM-files in path. Instances for the same pyramid
        level will be combined when possible to one file (e.g. not split
        for optical paths or focal planes). If instances are sparse tiled they
        will be converted to full tiled by inserting blank tiles. All instance uids will
        be changed.

        Parameters
        ----------
        output_path: Union[str, Path, UPath]
            Output folder to write files to. Should preferably be an dedicated folder
            for the wsi. Path can be local or an URL supported by fsspec.
        uid_generator: Callable[..., UID] = pydicom.uid.generate_uid
            Function that can generate unique identifiers.
        workers: Optional[int] = None
            Maximum number of thread workers to use.
        chunk_size: Optional[int] = None
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        offset_table: Optional[Union["str", OffsetTableType]] = None,
            Offset table to use, defined either by string (`empty`, `bot`, `eot`, or
            `none`) or `OffsetTableType` enum. Default to None, which will use
            `bot` for encapsulated  syntaxes and `none` for non-encapsulated transfer
            syntaxes.
        include_pyramids: Optional[Sequence[int]] = None
            Optional list of indices (in present pyramids) to include.
        include_levels: Optional[Sequence[int]] = None
            Optional list of indices (in all pyramids) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indicies can be used,
            e.g. [-1, -2] includes the two highest levels.
        include_labels: bool = True
            If to include label series.
        include_overviews: bool = True
            If to include overview series.
        add_missing_levels: bool = False
            If to add missing dyadic levels up to the single tile level.
        label: Optional[Union[Image, Union[str, Path, UPath]]] = None
            Optional label image to use instead of present label (if any).
        transcoding: Optional[Union[EncoderSettings, Encoder]] = None,
            Optional settings or encoder for transcoding image data. If None, image data
            will be copied as is.
        file_options: Optional[Dict[str, Any]] = None
            Optional options for saving files to output path.

        Returns
        -------
        List[UPath]
            List of paths of created files.
        """
        if workers is None:
            cpus = os.cpu_count()
            if cpus is None:
                workers = 1
            else:
                workers = cpus
        if chunk_size is None:
            chunk_size = 16
        if isinstance(offset_table, str):
            offset_table = OffsetTableType.from_string(offset_table)
        with WsiDicomFileTarget(
            output_path,
            uid_generator,
            workers,
            chunk_size,
            offset_table,
            include_pyramids,
            include_levels,
            add_missing_levels,
            transcoding,
            file_options,
        ) as target:
            target.save_pyramids(self.pyramids)
            if include_overviews and self.overviews is not None:
                target.save_overviews(self.overviews)
            if include_labels:
                if label is not None:
                    label_instances = [
                        WsiInstance.create_label(
                            label,
                            self._source.base_dataset,
                        )
                    ]
                    labels = Labels.open(label_instances)
                else:
                    labels = self.labels
                if labels is not None:
                    target.save_labels(labels)
            return target.filepaths

    @classmethod
    def is_ready_for_viewing(
        cls,
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]],
        file_options: Optional[Dict[str, Any]] = None,
    ) -> Optional[bool]:
        """
        Return true if files in path are formatted for fast viewing, i.e.
        have TILED_FULL tile arrangement and have an offset table.

        Parameters
        ----------
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]]
            Files to open. Can be a path for a single file, a list of paths for multiple
            files, or a path to a folder containing files. Path can be local or an URL
            supported by fsspec.
        file_options: Optional[Dict[str, Any]] = None
            Optional options for when opening files.

        Returns
            True if files in path are formatted for fast viewing, None if no DICOM WSI
            files are in the path.
        """
        source = WsiDicomFileSource.open(files, file_options)
        return source.is_ready_for_viewing

    @classmethod
    def is_supported(
        cls,
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]],
        file_options: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Return true if files in path have at least one level that can be read with
        WsiDicom.

        Parameters
        ----------
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]]
            Path to files to test. Path can be local or an URL supported by fsspec.
        file_options: Optional[Dict[str, Any]] = None
            Optional options for when opening files.

        Returns
            True if files in path have one level that can be read with WsiDicom.
        """
        source = WsiDicomFileSource.open(files, file_options)
        return source.contains_levels

    @property
    def size(self) -> Size:
        """Return pixel size of base level of selected pyramid."""
        return self.pyramids[self.selected_pyramid].size

    @property
    def mm_size(self) -> SizeMm:
        """Return image size in mm of selected pyramid."""
        return self.pyramids[self.selected_pyramid].mm_size

    @property
    def tile_size(self) -> Size:
        """Return tile size of selected pyramid."""
        return self.pyramids[self.selected_pyramid].tile_size

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm/pixel of base level of selected pyramid."""
        return self.pyramids[self.selected_pyramid].pixel_spacing

    @property
    def mpp(self) -> SizeMm:
        """Return pixel size in um/pixel of base level of selected pyramid."""
        return self.pyramids[self.selected_pyramid].mpp

    @property
    def selected_pyramid(self) -> int:
        """Return index of selected pyramid."""
        return self._selected_pyramid

    @property
    def uids(self) -> Optional[SlideUids]:
        return self._uids

    @property
    def pyramids(self) -> Pyramids:
        """Return contained pyramids."""
        if self._pyramids is not None:
            return self._pyramids
        raise WsiDicomNotFoundError("pyramids", str(self))

    @property
    def pyramid(self) -> Pyramid:
        """Return contained pyramid for selected pyramid."""
        return self.pyramids[self.selected_pyramid]

    @property
    def levels(self) -> Pyramid:
        """Return contained levels for selected pyramid."""
        return self.pyramid

    @property
    def labels(self) -> Optional[Labels]:
        """Return contained labels."""
        return self._labels

    @property
    def overviews(self) -> Optional[Overviews]:
        """Return contained overviews."""
        return self._overviews

    @property
    def annotations(self) -> List[AnnotationInstance]:
        """Return contained annotations."""
        return self._annotations

    @property
    def collection(self) -> List[Union[Pyramid, Labels, Overviews]]:
        collection: List[Optional[Union[Pyramid, Labels, Overviews]]] = [
            self._labels,
            self._overviews,
        ]
        collection.extend(pyramid for pyramid in self._pyramids)
        return [series for series in collection if series is not None]

    @property
    def metadata(self) -> WsiMetadata:
        return self._create_metadata(self._selected_pyramid)

    @property
    def files(self) -> Optional[List[UPath]]:
        """Return opened files if source is file-based."""
        if isinstance(self._source, WsiDicomFileSource):
            return self._source.files
        return None

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        string = self.__class__.__name__
        if depth is not None:
            depth -= 1
            if depth < 0:
                return string
        return (
            string
            + " of pyramids:\n"
            + list_pretty_str(list(self.pyramids), indent, depth, 0, 2)
        )

    def set_selected_pyramid(self, index: int) -> None:
        """Set selected pyramid.

        Parameters
        ----------
        index: int
            Index of pyramid to select.
        """
        if index < 0 or index >= len(self.pyramids):
            raise WsiDicomNotFoundError(f"Pyramid {index}", str(self.pyramids))
        self._selected_pyramid = index

    def read_label(self, index: int = 0) -> Image:
        """Read label image of the whole slide. If several label
        images are present, index can be used to select a specific image.

        Parameters
        ----------
        index: int = 0
            Index of the label image to read.

        Returns
        -------
        Image
            Label as Pillow image.
        """
        if self.labels is None:
            raise WsiDicomNotFoundError("label", str(self))
        label = self.labels.get(index)
        return label.get_default_full()

    def read_overview(self, index: int = 0) -> Image:
        """Read overview image of the whole slide. If several overview
        images are present, index can be used to select a specific image.

        Parameters
        ----------
        index: int = 0
            Index of the overview image to read.

        Returns
        -------
        Image
            Overview as Pillow image.
        """
        if self.overviews is None:
            raise WsiDicomNotFoundError("overview", str(self))
        overview = self.overviews.get(index)
        return overview.get_default_full()

    def read_thumbnail(
        self,
        size: Union[int, Tuple[int, int]] = 512,
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
    ) -> Image:
        """Read thumbnail image of the whole slide with dimensions no larger than given
        size.

        Parameters
        ----------
        size: Union[int, Tuple[int, int]] = 512
            Upper size limit for thumbnail.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read thumbnail from. If `None` the index in `selected_pyramid` is
            used.

        Returns
        -------
        Image
            Thumbnail as Pillow image,
        """
        if isinstance(size, int):
            size = (size, size)
        thumbnail_size = Size.from_tuple(size)
        if pyramid is None:
            pyramid = self.selected_pyramid
        level = self.pyramids.get(pyramid).get_closest_by_size(thumbnail_size)
        region = Region(position=Point(0, 0), size=level.size)
        image = level.get_region(region, z, path)
        image.thumbnail((size), resample=settings.pillow_resampling_filter)
        return image

    def read_region(
        self,
        location: Tuple[int, int],
        level: int,
        size: Tuple[int, int],
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
        threads: int = 1,
    ) -> Image:
        """Read region defined by pixels.

        Parameters
        ----------
        location: Tuple[int, int]
            Upper left corner of region in pixels.
        level: int
            Level in pyramid.
        size: Tuple[int, int]
            Size of region in pixels.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read region from. If `None` the index in `selected_pyramid` is
            used.
        threads: int = 1
            Number of threads to use for read.

        Returns
        -------
        Image
            Region as Pillow image.
        """
        if pyramid is None:
            pyramid = self.selected_pyramid
        wsi_level = self.pyramids.get(pyramid).get_closest_by_level(level)
        scale_factor = wsi_level.calculate_scale(level)
        scaled_region = (
            Region(position=Point.from_tuple(location), size=Size.from_tuple(size))
            * scale_factor
        )

        if not wsi_level.valid_pixels(scaled_region):
            raise WsiDicomOutOfBoundsError(
                f"Region {scaled_region}", f"level size {wsi_level.size}"
            )
        image = wsi_level.get_region(scaled_region, z, path, threads)
        if scale_factor != 1:
            image = image.resize((size), resample=settings.pillow_resampling_filter)
        return image

    def read_region_mm(
        self,
        location: Tuple[float, float],
        level: int,
        size: Tuple[float, float],
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
        slide_origin: bool = False,
        threads: int = 1,
    ) -> Image:
        """Read image from region defined in mm.

        Parameters
        ----------
        location: Tuple[float, float]
            Upper left corner of region in mm, or lower left corner if using slide
            origin.
        level: int
            Level in pyramid.
        size: Tuple[float, float].
            Size of region in mm.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read region from. If `None` the index in `selected_pyramid` is
            used.
        slide_origin: bool = False
            If to use the slide origin instead of image origin.
        threads: int = 1
            Number of threads to use for read.

        Returns
        -------
        Image
            Region as Pillow image.
        """
        if pyramid is None:
            pyramid = self.selected_pyramid
        wsi_level = self.pyramids.get(pyramid).get_closest_by_level(level)
        scale_factor = wsi_level.calculate_scale(level)
        region = RegionMm(PointMm.from_tuple(location), SizeMm.from_tuple(size))
        image = wsi_level.get_region_mm(region, z, path, slide_origin, threads)
        image_size = Size(width=image.size[0], height=image.size[1]) // scale_factor
        return image.resize(
            image_size.to_tuple(), resample=settings.pillow_resampling_filter
        )

    def read_region_mpp(
        self,
        location: Tuple[float, float],
        mpp: float,
        size: Tuple[float, float],
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
        slide_origin: bool = False,
        threads: int = 1,
    ) -> Image:
        """Read image from region defined in mm with set pixel spacing.

        Parameters
        ----------
        location: Tuple[float, float].
            Upper left corner of region in mm, or lower left corner if using slide
            origin.
        mpp: float
            Requested pixel spacing (um/pixel).
        size: Tuple[float, float].
            Size of region in mm.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read region from. If `None` the index in `selected_pyramid` is
            used.
        slide_origin: bool = False
            If to use the slide origin instead of image origin.
        threads: int = 1
            Number of threads to use for read.

        Returns
        --------
        Image
            Region as Pillow image.
        """
        pixel_spacing = mpp / 1000.0
        if pyramid is None:
            pyramid = self.selected_pyramid
        wsi_level = self.pyramids.get(pyramid).get_closest_by_pixel_spacing(
            SizeMm(pixel_spacing, pixel_spacing)
        )
        region = RegionMm(PointMm.from_tuple(location), SizeMm.from_tuple(size))
        image = wsi_level.get_region_mm(region, z, path, slide_origin, threads)
        image_size = SizeMm.from_tuple(size) // pixel_spacing
        return image.resize(
            image_size.to_tuple(), resample=settings.pillow_resampling_filter
        )

    def read_tile(
        self,
        level: int,
        tile: Tuple[int, int],
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
    ) -> Image:
        """Read tile in pyramid level as Pillow image.

        Parameters
        ----------
        level: int
            Pyramid level.
        tile: int, int
            tile xy coordinate.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read tile from. If `None` the index in `selected_pyramid` is
            used.

        Returns
        -------
        Image
            Tile as Pillow image.
        """
        tile_point = Point.from_tuple(tile)
        wsi_pyramid = self.pyramids.get(
            pyramid if pyramid is not None else self._selected_pyramid
        )
        try:
            wsi_level = wsi_pyramid.get(level)
            return wsi_level.get_tile(tile_point, z, path)
        except WsiDicomNotFoundError:
            # Scale from closest level
            wsi_level = wsi_pyramid.get_closest_by_level(level)
            return wsi_level.get_scaled_tile(tile_point, level, z, path)

    def read_encoded_tile(
        self,
        level: int,
        tile: Tuple[int, int],
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
    ) -> bytes:
        """Read tile in pyramid level as encoded bytes. For non-existing levels
        the tile is scaled down from a lower level, using the similar encoding.

        Parameters
        ----------
        level: int
            Pyramid level.
        tile: int, int
            Tile xy coordinate.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read tile from. If `None` the index in `selected_pyramid` is
            used.

        Returns
        -------
        bytes
            Tile in file encoding.
        """
        tile_point = Point.from_tuple(tile)
        wsi_pyramid = self.pyramids.get(
            pyramid if pyramid is not None else self._selected_pyramid
        )
        try:
            wsi_level = wsi_pyramid.get(level)
            return wsi_level.get_encoded_tile(tile_point, z, path)
        except WsiDicomNotFoundError:
            # Scale from closest level
            wsi_level = wsi_pyramid.get_closest_by_level(level)
            return wsi_level.get_scaled_encoded_tile(tile_point, level, z, path)

    def get_instance(
        self,
        level: int,
        z: Optional[float] = None,
        path: Optional[str] = None,
        pyramid: Optional[int] = None,
    ) -> WsiInstance:
        """Return instance fulfilling level, z and/or path.

        Parameters
        ----------
        level: int
            Pyramid level
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            Optical path, optional.
        pyramid: Optional[int] = None
            Pyramid to read tile from. If `None` the index in `selected_pyramid` is
            used.

        Returns
        -------
        WsiInstance:
            Instance
        """
        if pyramid is None:
            pyramid = self.selected_pyramid
        wsi_level = self.pyramids.get(pyramid).get(level)
        return wsi_level.get_instance(z, path)

    def close(self) -> None:
        """Close source if owned by this instance."""
        if self._source_owned:
            self._source.close()

    def clear_cache(self):
        """Clear cache of encoded and decoded tiles."""
        self._source.clear_cache()

    def resize_cache(self, size: int):
        """Resize cache of encoded and decoded tiles."""
        self._source.resize_cache(size)

    def _validate_collection(self) -> SlideUids:
        """
        Check that no files or instance in collection is duplicate, and, if
        strict, that all series have the same base uids.
        Raises WsiDicomMatchError otherwise. Returns base uid for collection.

        Returns
        -------
        SlideUids
            Matching uids
        """
        datasets = [
            dataset for collection in self.collection for dataset in collection.datasets
        ]
        WsiDataset.check_duplicate_dataset(datasets, self)

        instances = [
            instance
            for collection in self.collection
            for instance in collection.instances
        ]

        WsiInstance.check_duplicate_instance(instances, self)

        slide_uids = self.levels.uids
        if slide_uids is None:
            raise WsiDicomNotFoundError("Valid levels", str(self))
        for series in self.collection:
            if series.uids is not None and not series.uids.matches(slide_uids):
                raise WsiDicomMatchError(str(series), str(self))

        if self.annotations != []:
            for annotation in self.annotations:
                if not annotation.slide_uids.matches(slide_uids):
                    logging.warning("Annotations uids does not match.")
        return slide_uids

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.pyramids}, {self.labels}"
            f"{self.overviews}, {self.annotations})"
        )

    def __str__(self) -> str:
        return self.pretty_str()

    @lru_cache
    def _create_metadata(self, pyramid_index: int) -> WsiMetadata:
        if self.labels is not None:
            label_metadata = self.labels.metadata
        else:
            label_metadata = None
        if self.overviews is not None:
            overview_metadata = self.overviews.metadata
        else:
            overview_metadata = None
        pyramid = self.pyramids.get(pyramid_index)
        return WsiMetadata.merge_image_types(
            pyramid.metadata, label_metadata, overview_metadata
        )



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from collections import defaultdict
from functools import cached_property
from typing import Dict, List, Optional, Sequence, Set, Tuple

import numpy as np

from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.geometry import Point, Size
from wsidicom.instance.dataset import WsiDataset
from wsidicom.instance.tile_index.tile_index import TileIndex


class SparseTilePlane:
    """Hold frame indices for the tiles in a sparse tiled file. Empty (sparse)
    frames are represented by -1."""

    def __init__(self, tiled_size: Size):
        """Create a SparseTilePlane of specified size.

        Parameters
        ----------
        tiled_size: Size
            Size of the tiling
        """
        self._shape = tiled_size
        self.plane = np.full(tiled_size.to_tuple(), -1, dtype=np.dtype(int))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._shape})"

    def __str__(self) -> str:
        return self.pretty_str()

    def __getitem__(self, position: Point) -> int:
        """Get frame index from tile index at plane_position.

        Parameters
        ----------
        plane_position: Point
            Position in plane to get the frame index from

        Returns
        -------
        int
            Frame index
        """
        frame_index = int(self.plane[position.x, position.y])
        return frame_index

    def __setitem__(self, position: Point, frame_index: int):
        """Add frame index to tile index.

        Parameters
        ----------
        plane_position: Point
            Position in plane to add the frame index
        frame_index: int
            Frame index to add to the index
        """
        self.plane[position.x, position.y] = frame_index

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        return "Sparse tile plane"


class SparseTileIndex(TileIndex):
    """Index for mapping tile position to frame number for datasets containing
    sparse tiles. Frame indices are retrieved from tile position, z, and path
    by finding the corresponding matching SparseTilePlane (z and path) and
    returning the frame index at tile position. If the tile is missing (due to
    the sparseness), -1 is returned."""

    def __init__(self, datasets: Sequence[WsiDataset]):
        """Create sparse tile index for frames in datasets. Requires equal tile
        size for all tile planes. Pixel data tiles are identified by the Per
        Frame Functional Groups Sequence that contains tile column, row, z,
        path, and frame index. These are stored in a SparseTilePlane
        (one plane for every combination of z and path).

        Parameters
        ----------
        datasets: Sequence[Dataset]
            List of datasets containing sparse tiled image data.
        """
        super().__init__(datasets)

    @cached_property
    def planes(self) -> Dict[Tuple[float, str], SparseTilePlane]:
        return self._read_planes_from_datasets()

    @cached_property
    def focal_planes(self) -> List[float]:
        return self._get_focal_planes()

    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        return (
            f"Sparse tile index tile size: {self.tile_size}, "
            f"plane size: {self.tiled_size}"
        )

    def get_frame_index(self, tile: Point, z: float, path: str) -> int:
        """Return frame index for a Point tile, z coordinate, and optical
        path.

        Parameters
        ----------
        tile: Point
            Tile xy to get.
        z: float
            Z coordinate to get.
        path: str
            ID of optical path to get.

        Returns
        -------
        int
            Frame index.
        """
        try:
            plane = self.planes[(z, path)]
        except KeyError:
            raise WsiDicomNotFoundError(f"Plane with z {z}, path {path}", str(self))
        frame_index = plane[tile]
        return frame_index

    def _get_focal_planes(self) -> List[float]:
        """Return list of focal planes defined in planes.

        Returns
        -------
        List[float]
            Focal planes, specified in um.
        """
        focal_planes: Set[float] = set()
        for z, _ in self.planes.keys():
            focal_planes.add(z)
        return sorted(list(focal_planes))

    def _read_planes_from_datasets(self) -> Dict[Tuple[float, str], SparseTilePlane]:
        """Return SparseTilePlane from planes in datasets.

        Parameters
        ----------
        datasets: Sequence[WsiDataset]
           List of datasets to read planes from.

        Returns
        -------
        Dict[Tuple[float, str], SparseTilePlane]
            Dict of planes with focal plane and optical identifier as key.
        """
        planes: Dict[Tuple[float, str], SparseTilePlane] = defaultdict(
            lambda: SparseTilePlane(self.tiled_size)
        )
        for dataset in self._datasets:
            frame_sequence = dataset.frame_sequence
            for i, frame in enumerate(frame_sequence):
                (tile, z) = self._read_frame_coordinates(frame)
                identifier = dataset.read_optical_path_identifier(frame)
                planes[(z, identifier)][tile] = i + dataset.frame_offset
        return planes



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with settings for encoding image data."""

from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Literal, Optional, Sequence, Union

from pydicom.uid import (
    JPEG2000,
    UID,
    ExplicitVRBigEndian,
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    JPEG2000Lossless,
    JPEGBaseline8Bit,
    JPEGExtended12Bit,
    JPEGLossless,
    JPEGLosslessSV1,
    JPEGLSLossless,
    JPEGLSNearLossless,
    RLELossless,
)


class Channels(Enum):
    """Color channels in encoded image."""

    GRAYSCALE = "grayscale"
    YBR = "ybr"
    RGB = "rgb"

    @classmethod
    def from_photometric_interpretation(
        cls, photometric_interpretation: str
    ) -> "Channels":
        """Return channels matching photometric interpretation."""
        ybr_photometric_interpretations = [
            "YBR_FULL",
            "YBR_FULL_422",
            "YBR_PARTIAL_422",
            "YBR_ICT",
            "YBR_RCT",
            "YBR",
        ]
        if photometric_interpretation == "MONOCHROME2":
            return cls.GRAYSCALE
        if photometric_interpretation in ybr_photometric_interpretations:
            return cls.YBR
        if photometric_interpretation == "RGB":
            return cls.RGB
        raise ValueError(
            f"Unsupported photometric interpretation: {photometric_interpretation}."
        )


class Subsampling(Enum):
    """JPEG chroma subsampling settings."""

    R444 = "Full horizontal, full vertical (4:4:4)"
    R422 = "Half horisontal, full vertical (4:2:2)"
    R420 = "Half horisontal, half vertical (4:2:0)"
    R411 = "Quarter horisontal, full vertical (4:1:1)"
    R440 = "Full horizontal, half vertical (4:4:0)"

    @classmethod
    def from_string(cls, string: Optional[str]) -> "Subsampling":
        """Return subsampling matching string."""
        if string is None or string == "444":
            return Subsampling.R444
        if string == "422":
            return Subsampling.R422
        if string == "420":
            return Subsampling.R420
        if string == "411":
            return Subsampling.R411
        if string == "440":
            return Subsampling.R440
        raise ValueError(f"Unsupported subsampling: {string}.")


class Settings(metaclass=ABCMeta):
    def __init__(self, bits: int, channels: Channels):
        """
        Initialize encoding settings.

        Parameters:
        ----------
            bits: int
                Number of bits per pixel.
            channels: Channels
                Color channels in encoded image."""
        self._bits = bits
        self._channels = channels

    @property
    def bits(self) -> int:
        """Return bits."""
        return self._bits

    @property
    def channels(self) -> Channels:
        """Return channels."""
        return self._channels

    @property
    @abstractmethod
    def transfer_syntax(self) -> UID:
        """Return transfer syntax."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def extension(self) -> str:
        """Return file extension."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def photometric_interpretation(self) -> str:
        """Return photometric interpretation."""
        raise NotImplementedError()

    @property
    def allocated_bits(self) -> int:
        """Return allocated bits."""
        return self.bits // 8 * 8

    @property
    def high_bit(self) -> int:
        """Return high bit."""
        return self.bits - 1

    @property
    def samples_per_pixel(self) -> int:
        """Return samples per pixel."""
        return 1 if self.channels == Channels.GRAYSCALE else 3

    @classmethod
    def create(
        cls,
        transfer_syntax: UID,
        bits: int,
        photometric_interpretation: str,
    ) -> "Settings":
        """Create settings based on properties dataset and transfer syntax.

        Parameters:
        ----------
            transfer_syntax: UID
                Transfer syntax to create settings for.
            bits: int
                Number of bits per pixel.
            photometric_interpretation: str
                Photometric interpretation of image.

        Returns:
        ----------
            Settings
                Settings matching dataset and transfer syntax.
        """
        jpeg_transfer_syntaxes = [
            JPEGBaseline8Bit,
            JPEGExtended12Bit,
        ]
        jpeg_lossless_transfer_syntaxes = [
            JPEGLossless,
            JPEGLosslessSV1,
        ]
        jpeg_ls_transfer_syntaxes = [
            JPEGLSLossless,
            JPEGLSNearLossless,
        ]
        jpeg_2000_transfer_syntaxes = [
            JPEG2000Lossless,
            JPEG2000,
        ]
        uncompressed_transfer_syntaxes = [
            ImplicitVRLittleEndian,
            ExplicitVRLittleEndian,
            ExplicitVRBigEndian,
        ]
        channels = Channels.from_photometric_interpretation(photometric_interpretation)
        if transfer_syntax in jpeg_transfer_syntaxes:
            return JpegSettings(bits=bits, channels=channels)
        if transfer_syntax in jpeg_lossless_transfer_syntaxes:
            return JpegLosslessSettings(bits=bits, channels=channels)
        if transfer_syntax in jpeg_ls_transfer_syntaxes:
            if channels == Channels.YBR:
                raise ValueError("JPEG-LS does not support conversion to ybr.")
            if transfer_syntax == JPEGLSNearLossless:
                return JpegLsSettings(level=1, bits=bits, channels=channels)
            return JpegLsSettings(level=0, bits=bits, channels=channels)
        if transfer_syntax in jpeg_2000_transfer_syntaxes:
            if transfer_syntax == JPEG2000:
                return Jpeg2kSettings(bits=bits, channels=channels)
            return Jpeg2kSettings(levels=0, bits=bits, channels=channels)
        if transfer_syntax == RLELossless:
            return RleSettings(bits=bits, channels=channels)
        if transfer_syntax in uncompressed_transfer_syntaxes:
            if channels == Channels.YBR:
                raise ValueError("Numpy encoder does not support ybr images.")
            return NumpySettings(
                bits=bits,
                channels=channels,
                little_endian=transfer_syntax.is_little_endian,
                implicit_vr=transfer_syntax.is_implicit_VR,
            )
        raise ValueError(f"Unsupported transfer syntax: {transfer_syntax}.")


class JpegSettings(Settings):
    """JPEG encoding settings."""

    def __init__(
        self,
        quality: int = 80,
        bits: int = 8,
        channels: Channels = Channels.YBR,
        subsampling: Subsampling = Subsampling.R422,
    ):
        """
        Initialize JPEG encoding settings.

        Parameters:
        ----------
            quality: int = 80
                JPEG quality factor. 0-100, recommended to not use higher than 95.
            bits: int
                Number of bits per pixel.
            channels: Channels = Channels.YBR
                Color channels in encoded image.
            subsampling: Subsampling = Subsampling.R420
                Chroma subsampling settings.
        """
        self._quality = quality
        self._subsampling = subsampling
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(quality={self.quality}, bits={self.bits}, "
            f"channels={self.channels}, subsampling={self.subsampling})"
        )

    @property
    def quality(self) -> int:
        """Return quality."""
        return self._quality

    @property
    def subsampling(self) -> Subsampling:
        """Return subsampling."""
        return self._subsampling

    @property
    def transfer_syntax(self) -> UID:
        if self.bits == 8:
            return JPEGBaseline8Bit
        return JPEGExtended12Bit

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            return "YBR_FULL_422"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jpg"


class JpegLosslessSettings(Settings):
    def __init__(
        self,
        predictor: int = 1,
        bits: int = 8,
        channels: Channels = Channels.YBR,
    ):
        """
        Initialize JPEG lossless encoding settings.

        Parameters:
        ----------
            predictor: int = 1
                JPEG predictor setting. 1-7. Use 1 for SV1.
            bits: int = 8
                Number of bits per pixel.
            channels: Channels = Channels.YBR
                Color channels in encoded image.
        """
        self._predictor = predictor
        super().__init__(bits, channels)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(predictor={self.predictor}, bits={self.bits}, "
            f"channels={self.channels})"
        )

    @property
    def predictor(self) -> int:
        """Return predictor."""
        return self._predictor

    @property
    def transfer_syntax(self) -> UID:
        if self.predictor is None:
            return JPEGLosslessSV1
        return JPEGLossless

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            return "YBR_FULL"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jpg"


class JpegLsSettings(Settings):
    def __init__(
        self,
        level: int = 0,
        bits: int = 8,
        channels: Channels = Channels.RGB,
    ):
        """
        Initialize JPEG-LS lossless encoding settings. Only supports grayscale images.

        Parameters:
        ----------
            level: int = 0
                JPEG-LS near-lossless level. 0-9. 0 is lossless.
            bits: int = 8
                Number of bits per pixel.
        """
        self._level = level
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(level={self.level}, bits={self.bits})"

    @property
    def level(self) -> int:
        """Return level."""
        return self._level

    @property
    def transfer_syntax(self) -> UID:
        if self.level == 0:
            return JPEGLSLossless
        return JPEGLSNearLossless

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jls"


class Jpeg2kSettings(Settings):
    def __init__(
        self,
        levels: Union[int, Sequence[int]] = 80,
        bits: int = 8,
        channels: Channels = Channels.YBR,
    ):
        """
        Initialize JPEG 2000 encoding settings.

        Parameters:
        ----------
            levels: Union[float, Sequence[float]] = 80.0
                JPEG 2000 compression levels in dB. Set to 0.
            bits: int = 8
                Number of bits per pixel.
            channels: Channels = Channels.YBR
                Color channels in encoded image.
        """
        if isinstance(levels, int):
            levels = [levels]
        self._levels = levels
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(levels={self.levels}, bits={self.bits}, "
            f"channels={self.channels})"
        )

    @property
    def levels(self) -> Sequence[int]:
        """Return level."""
        return self._levels

    @property
    def lossless(self) -> bool:
        """Return True if lossless, else False."""
        return self.levels[-1] == 0

    @property
    def transfer_syntax(self) -> UID:
        if self.lossless:
            return JPEG2000Lossless
        return JPEG2000

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            if self.lossless:
                return "YBR_RCT"
            return "YBR_ICT"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".jp2"


class NumpySettings(Settings):
    def __init__(
        self,
        bits: int = 8,
        channels: Union[
            Literal[Channels.GRAYSCALE], Literal[Channels.RGB]
        ] = Channels.RGB,
        little_endian: bool = True,
        implicit_vr: bool = False,
    ):
        """
        Initialize numpy encoding settings.

        Parameters:
        ----------
            bits: int = 8
                Number of bits per pixel.
            channels: Union[
                Literal[Channels.GRAYSCALE], Literal[Channels.RGB]
            ] = Channels.RGB,
                Color channels in encoded image.
            little_endian: bool = True
                Endianness of encoded image.
            implicit_vr: bool = False
                VR encoding of transfer syntax.
        """
        self._little_endian = little_endian
        self._implicit_vr = implicit_vr
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(bits={self.bits}, channels={self.channels}, "
            f"little_endian={self.little_endian}, implicit_vr={self.implicit_vr})"
        )

    @property
    def little_endian(self) -> bool:
        """Return little endian."""
        return self._little_endian

    @property
    def implicit_vr(self) -> bool:
        """Return implicit vr."""
        return self._implicit_vr

    @property
    def transfer_syntax(self) -> UID:
        if self.little_endian:
            if self.implicit_vr:
                return ImplicitVRLittleEndian
            return ExplicitVRLittleEndian
        return ExplicitVRBigEndian

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".raw"


class RleSettings(Settings):
    def __init__(self, bits: int = 8, channels: Channels = Channels.RGB):
        super().__init__(bits, channels)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bits={self.bits}, channels={self.channels})"

    @property
    def transfer_syntax(self) -> UID:
        return RLELossless

    @property
    def photometric_interpretation(self) -> str:
        if self.channels == Channels.GRAYSCALE:
            return "MONOCHROME2"
        if self.channels == Channels.YBR:
            return "YBR_FULL"
        return "RGB"

    @property
    def extension(self) -> str:
        return ".rle"



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with encoders for image data."""

import io
from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Dict, Generic, Optional, Tuple, Type, TypeVar, Union

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom import Dataset, FileMetaDataset
from pydicom.pixels.utils import pixel_dtype
from pydicom.uid import (
    JPEG2000,
    UID,
    JPEGBaseline8Bit,
    JPEGExtended12Bit,
    JPEGLSNearLossless,
)

from wsidicom.codec.optionals import (
    IMAGE_CODECS_AVAILABLE,
    JPEG8,
    PYLIBJPEGLS_AVAILABLE,
    PYLIBJPEGOPENJPEG_AVAILABLE,
    PYLIBJPEGRLE_AVAILABLE,
    jpeg2k_encode,
    jpeg8_encode,
    jpegls_encode,
    pylibjpeg_ls_encode,
    pylibjpeg_openjpeg_encode,
    rle_encode_frame,
)
from wsidicom.codec.rle import RleCodec
from wsidicom.codec.settings import (
    Channels,
    Jpeg2kSettings,
    JpegLosslessSettings,
    JpegLsSettings,
    JpegSettings,
    NumpySettings,
    RleSettings,
    Settings,
    Subsampling,
)
from wsidicom.uid import HTJPEG2000

SettingsType = TypeVar("SettingsType", bound=Settings)


class LossyCompressionIsoStandard(Enum):
    JPEG_LOSSY = "ISO_10918_1"
    JPEG_LS_NEAR_LOSSLESS = "ISO_14495_1"
    JPEG_2000_IRREVERSIBLE = "ISO_15444_1"
    HT_JPEG_2000_IRREVERSIBLE = "ISO_15444_15"

    @classmethod
    def transfer_syntax_to_iso(
        cls, transfer_syntax: UID
    ) -> Optional["LossyCompressionIsoStandard"]:
        if transfer_syntax in [JPEGBaseline8Bit, JPEGExtended12Bit]:
            return cls.JPEG_LOSSY
        elif transfer_syntax == JPEGLSNearLossless:
            return cls.JPEG_LS_NEAR_LOSSLESS
        elif transfer_syntax == JPEG2000:
            return cls.JPEG_2000_IRREVERSIBLE
        elif transfer_syntax == HTJPEG2000:
            return cls.HT_JPEG_2000_IRREVERSIBLE
        return None


class Encoder(Generic[SettingsType], metaclass=ABCMeta):
    """Abstract base class for encoders."""

    def __init__(self, settings: SettingsType):
        """Initialize encoder.

        Parameters
        ----------
        settings: Settings
            Settings for the encoder.
        """
        self._settings = settings

    @abstractmethod
    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        """Encode image into bytes.

        Parameters
        ----------
        image: Union[Image, np.ndarray]
            Image to encode.

        Returns
        -------
        bytes
            Encoded image.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def lossy(self) -> bool:
        """Return True if encoder is lossy."""
        raise NotImplementedError()

    @property
    def lossy_method(self) -> Optional[LossyCompressionIsoStandard]:
        """Return ISO standard name of compression if encoder is lossy."""
        return LossyCompressionIsoStandard.transfer_syntax_to_iso(self.transfer_syntax)

    @classmethod
    @abstractmethod
    def supports_settings(cls, settings: Settings) -> bool:
        """Return True if encoder supports settings."""
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        """Return True if encoder is available."""
        raise NotImplementedError()

    @property
    def settings(self) -> SettingsType:
        """Return the settings."""
        return self._settings

    @property
    def transfer_syntax(self) -> UID:
        """Return the transfer syntax UID."""
        return self._settings.transfer_syntax

    @property
    def bits(self) -> int:
        """Return the number of bits per sample."""
        return self._settings.bits

    @property
    def allocated_bits(self) -> int:
        """Return the number of allocated bits."""
        return self._settings.allocated_bits

    @property
    def high_bit(self) -> int:
        """Return the high bit."""
        return self._settings.high_bit

    @property
    def channels(self) -> Channels:
        """Return the number of channels."""
        return self._settings.channels

    @property
    def samples_per_pixel(self) -> int:
        """Return the number of samples per pixel."""
        return self._settings.samples_per_pixel

    @property
    def photometric_interpretation(self) -> str:
        """Return the photometric interpretation."""
        return self._settings.photometric_interpretation

    @classmethod
    def create(
        cls, transfer_syntax: UID, bits: int, photometric_interpretation: str
    ) -> "Encoder":
        """Create an encoder using settings.

        Parameters
        ----------
        transfer_syntax: UID
            Transfer syntax UID.
        bits: int
            Number of bits per sample.
        photometric_interpretation: str
            Photometric interpretation.

        Returns
        -------
        Encoder
            Encoder for settings.
        """
        settings = Settings.create(
            transfer_syntax,
            bits,
            photometric_interpretation,
        )
        return cls.create_for_settings(settings)

    @classmethod
    def create_for_settings(cls, settings: Settings) -> "Encoder":
        """Create an encoder using settings.

        Parameters
        ----------
        settings: Settings
            Settings for the encoder.

        Returns
        -------
        Encoder
            Encoder for settings.
        """
        encoder = cls._select_encoder(settings)
        if encoder is None:
            raise ValueError(f"Unsupported encoder settings: {settings}")
        return encoder(settings)

    @staticmethod
    def _select_encoder(
        settings: Settings,
    ) -> Optional["Type[Encoder]"]:
        """Select encoder for settings.

        Parameters
        ----------
        settings: Settings
            Settings for the encoder.

        Returns
        -------
        Optional[Type[Encoder]]
            Encoder for settings, or None if no encoder is available.
        """
        # Sort encoders by preference
        encoders_by_settings: Dict[Type[Settings], Tuple[Type[Encoder], ...]] = {
            JpegSettings: (JpegEncoder, PillowEncoder),
            JpegLosslessSettings: (JpegEncoder,),
            JpegLsSettings: (JpegLsEncoder, PyJpegLsEncoder),
            Jpeg2kSettings: (Jpeg2kEncoder, PyLibJpegOpenJpegEncoder, PillowEncoder),
            NumpySettings: (NumpyEncoder,),
            RleSettings: (PylibjpegRleEncoder, ImageCodecsRleEncoder),
        }
        return next(
            (
                encoder
                for encoder in encoders_by_settings[type(settings)]
                if encoder.is_available() and encoder.supports_settings(settings)
            ),
            None,
        )


class PillowEncoder(Encoder[Union[JpegSettings, Jpeg2kSettings]]):
    """Encoder that uses pillow to encode image."""

    _supported_subsamplings = {
        Subsampling.R444: 0,
        Subsampling.R422: 1,
        Subsampling.R420: 2,
    }

    def __init__(self, settings: Union[JpegSettings, Jpeg2kSettings]):
        if isinstance(settings, JpegSettings):
            if settings.subsampling not in self._supported_subsamplings:
                raise ValueError(f"Unsupported subsampling: {settings.subsampling}.")
            subsampling = self._supported_subsamplings[settings.subsampling]
            self._pillow_settings = {
                "quality": settings.quality,
                "subsampling": subsampling,
            }
            self._format = "jpeg"
        elif isinstance(settings, Jpeg2kSettings):
            if len(settings.levels) != 1:
                raise ValueError("Only one level is supported.")
            self._pillow_settings = {
                "quality_mode": "dB",
                "quality_layers": [settings.levels[0]],
                "irreversible": not settings.lossless,
                "mct": settings.channels == Channels.YBR,
                "no_jp2": True,
            }
            self._format = "jpeg2000"
        else:
            raise ValueError(f"Unsupported encoder settings: {type(settings)}.")
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return isinstance(self.settings, JpegSettings) or (
            isinstance(self.settings, Jpeg2kSettings) and not self.settings.lossless
        )

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if not isinstance(image, Image):
            image = Pillow.fromarray(image)
        with io.BytesIO() as buffer:
            image.save(
                buffer, format=self._format, **self._pillow_settings  # type: ignore
            )
            return buffer.getvalue()

    @classmethod
    def is_available(cls) -> bool:
        return True

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        if isinstance(settings, JpegSettings):
            return (
                settings.bits == 8
                and settings.subsampling in cls._supported_subsamplings
            )
        return (
            isinstance(settings, Jpeg2kSettings)
            and settings.bits == 8
            and len(settings.levels) == 1
        )


class JpegEncoder(Encoder[Union[JpegSettings, JpegLosslessSettings]]):
    """JPEG encoder using image codecs."""

    def __init__(
        self,
        settings: Union[JpegSettings, JpegLosslessSettings],
    ) -> None:
        """Initialize JPEG encoder.

        Parameters
        ----------
        settings: Union[JpegSettings, JpegLosslessSettings]
            Settings for the encoder.

        """
        self._bits = settings.bits
        if self._bits == 8:
            self._dtype = np.uint8
        else:
            self._dtype = np.uint16
        if settings.channels == Channels.GRAYSCALE:
            self._output_colorspace = JPEG8.CS.GRAYSCALE
        elif settings.channels == Channels.YBR:
            self._output_colorspace = JPEG8.CS.YCbCr
        elif settings.channels == Channels.RGB:
            self._output_colorspace = JPEG8.CS.RGB
        else:
            raise ValueError(f"Unsupported channels: {settings.channels}.")
        if isinstance(settings, JpegSettings):
            self._lossless = False
            self._predictor = None
            self._level = settings.quality

            self._subsampling = (
                settings.subsampling.name.lstrip("R") if settings.subsampling else None
            )
        elif isinstance(settings, JpegLosslessSettings):
            self._lossless = True
            self._predictor = settings.predictor
            self._level = None
            self._subsampling = None
        else:
            raise ValueError(f"Unsupported encoder settings: {type(settings)}.")
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return not self._lossless

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if not self.is_available():
            raise RuntimeError("Image codecs not available.")
        return jpeg8_encode(
            np.array(image).astype(self._dtype),
            level=self._level,
            lossless=self._lossless,
            bitspersample=self._bits,
            subsampling=self._subsampling,
            outcolorspace=self._output_colorspace,
            predictor=self._predictor,
        )

    @classmethod
    def is_available(cls) -> bool:
        return IMAGE_CODECS_AVAILABLE

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, (JpegSettings, JpegLosslessSettings))


class JpegLsEncoder(Encoder[JpegLsSettings]):
    """Encoder that uses jpegls to encode image."""

    def __init__(
        self,
        settings: JpegLsSettings,
    ) -> None:
        """Initialize JPEG-LS encoder.

        Parameters
        ----------
        settings: JpegLsSettings
            Settings for the encoder.
        """
        self._bits = settings.bits
        if settings.level is None or settings.level == 0:
            self._level = 0
        else:
            self._level = settings.level
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return self._level != 0

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        """Encode image into bytes."""
        if not self.is_available():
            raise RuntimeError("Image codecs not available.")
        if isinstance(image, Image) and image.mode not in ("L", "RGB", "I;16", "I;16L"):
            raise ValueError(f"Unsupported mode: {image.mode}.")
        return jpegls_encode(np.array(image), level=self._level)

    @classmethod
    def is_available(cls) -> bool:
        return IMAGE_CODECS_AVAILABLE

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, JpegLsSettings)


class Jpeg2kEncoder(Encoder[Jpeg2kSettings]):
    """Encoder that uses jpeg2k to encode image."""

    def __init__(self, settings: Jpeg2kSettings) -> None:
        """Initialize JPEG 2000 encoder.

        Parameters
        ----------
        settings: Jpeg2kSettings
            Settings for the encoder.
        """
        if len(settings.levels) != 1:
            raise ValueError("Only one level is supported.")
        self._bits = settings.bits
        if settings.channels == Channels.YBR:
            self._multiple_component_transform = True
        else:
            self._multiple_component_transform = False
        if settings.lossless:
            self._level = 0
            self._reversible = True
        else:
            self._level = settings.levels[0]
            self._reversible = False
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return not self.settings.lossless

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if not self.is_available():
            raise RuntimeError("Image codecs not available.")
        if isinstance(image, Image) and image.mode not in ("L", "RGB", "I;16", "I;16L"):
            raise ValueError(f"Unsupported mode: {image.mode}.")
        return jpeg2k_encode(
            np.array(image),
            level=self._level,
            reversible=self._reversible,
            bitspersample=self._bits,
            codecformat="J2K",
            mct=self._multiple_component_transform,
        )

    @classmethod
    def is_available(cls) -> bool:
        return IMAGE_CODECS_AVAILABLE

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, Jpeg2kSettings) and len(settings.levels) == 1


class NumpyEncoder(Encoder[NumpySettings]):
    """Encoder that uses numpy to encode image."""

    def __init__(self, settings: NumpySettings) -> None:
        """Initialize numpy encoder.

        Parameters
        ----------
        settings: NumpySettings
            Settings for the encoder.
        """
        dataset = Dataset()
        dataset.BitsAllocated = settings.allocated_bits
        dataset.PixelRepresentation = 0
        dataset.file_meta = FileMetaDataset()
        dataset.file_meta.TransferSyntaxUID = settings.transfer_syntax
        self._dtype = pixel_dtype(dataset)
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return False

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if isinstance(image, Image) and image.mode not in ("L", "RGB", "I;16", "I;16L"):
            raise ValueError(f"Unsupported mode: {image.mode}.")
        return np.array(image).astype(self._dtype).tobytes()

    @classmethod
    def is_available(cls) -> bool:
        return True

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, NumpySettings)


class RleEncoder(Encoder[RleSettings]):
    def __init__(self, settings: RleSettings) -> None:
        """Initialize RLE encoder.

        Parameters
        ----------
        settings: RleSettings
            Settings for the encoder.
        """
        self._bits = settings.bits
        if settings.bits == 8:
            self._dtype = np.uint8
        else:
            self._dtype = np.uint16
        if settings.channels == Channels.GRAYSCALE:
            self._samples_per_pixel = 1
        else:
            self._samples_per_pixel = 3
        super().__init__(settings)

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if isinstance(image, Image) and image.mode not in ("L", "RGB", "I;16", "I;16L"):
            raise ValueError(f"Unsupported mode: {image.mode}.")
        return self._encode(np.array(image).astype(self._dtype))

    @property
    def lossy(self) -> bool:
        return False

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, RleSettings)

    @abstractmethod
    def _encode(self, image: np.ndarray) -> bytes:
        raise NotImplementedError()


class ImageCodecsRleEncoder(RleEncoder):
    """Encoder that uses image codecs PackBits to encode image."""

    def _encode(self, image: np.ndarray):
        if not self.is_available():
            raise RuntimeError("Image codecs not available.")
        return RleCodec.encode(image)

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        if settings.bits != 8:
            return False
        return isinstance(settings, RleSettings)

    @classmethod
    def is_available(cls) -> bool:
        return IMAGE_CODECS_AVAILABLE


class PylibjpegRleEncoder(RleEncoder):
    """Encoder that uses pylibjpeg-rle to encode image."""

    def _encode(self, image: np.ndarray) -> bytes:
        if not self.is_available():
            raise RuntimeError("Pylibjpeg-rle not available.")
        rows, cols = image.shape[0:2]
        return rle_encode_frame(
            np.array(image).astype(self._dtype).tobytes(),
            rows=rows,
            cols=cols,
            bpp=self._bits,
            spp=self._samples_per_pixel,
            byteorder="<",
        )

    @classmethod
    def is_available(cls) -> bool:
        return PYLIBJPEGRLE_AVAILABLE


class PyJpegLsEncoder(Encoder[JpegLsSettings]):
    """Encoder that uses pylibjpeg-ls to encode image."""

    def __init__(
        self,
        settings: JpegLsSettings,
    ) -> None:
        """Initialize PyJpegLs (JPEG LS) encoder.

        Parameters
        ----------
        settings: JpegLsSettings
            Settings for the encoder.
        """
        if settings.channels == Channels.GRAYSCALE:
            self._interleave_mode = 0
        else:
            self._interleave_mode = 2
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return False

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if not self.is_available():
            raise RuntimeError("Pylibjpeg-ls not available.")
        return pylibjpeg_ls_encode(
            np.asarray(image),
            lossy_error=self.settings.level,
            interleave_mode=self._interleave_mode,
        )

    @classmethod
    def is_available(cls) -> bool:
        return PYLIBJPEGLS_AVAILABLE

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, JpegLsSettings)


class PyLibJpegOpenJpegEncoder(Encoder[Jpeg2kSettings]):
    """Encoder that uses pylibjpeg-openjpeg to encode image."""

    def __init__(self, settings: Jpeg2kSettings) -> None:
        """Initialize PyLibjpeg OpenJpeg (JPEG 2000) encoder.

        Parameters
        ----------
        settings: Jpeg2kSettings
            Settings for the encoder.
        """
        self._bits = settings.bits
        if settings.channels == Channels.YBR:
            self._multiple_component_transform = True
        else:
            self._multiple_component_transform = False
        if settings.channels == Channels.GRAYSCALE:
            self._color_space = 2
        else:
            self._color_space = 1
        if settings.lossless:
            self._levels = None
        else:
            self._levels = settings.levels
        super().__init__(settings)

    @property
    def lossy(self) -> bool:
        return not self.settings.lossless

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        if not self.is_available():
            raise RuntimeError("Pylibjpeg-openjpeg not available.")
        return pylibjpeg_openjpeg_encode(
            np.asarray(image),
            self.bits,
            self._color_space,
            self._multiple_component_transform,
            signal_noise_ratios=list(self._levels) if self._levels else None,
        )

    @classmethod
    def is_available(cls) -> bool:
        return PYLIBJPEGOPENJPEG_AVAILABLE

    @classmethod
    def supports_settings(cls, settings: Settings) -> bool:
        return isinstance(settings, Jpeg2kSettings)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with default values for DICOM serialization."""

import datetime

from wsidicom.conceptcode import (
    ContainerComponentTypeCode,
    ContainerTypeCode,
    IlluminationCode,
    IlluminationColorCode,
)
from wsidicom.geometry import SizeMm
from wsidicom.metadata.image import FocusMethod


class Defaults:
    string = "Unknown"
    date_time = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)
    optical_path_identifier = "0"
    illumination_type = IlluminationCode("Brightfield illumination")
    illumination = IlluminationColorCode("Full Spectrum")
    slide_container_type = ContainerTypeCode("Microscope slide")
    slide_component_type = ContainerComponentTypeCode("Microscope slide cover slip")
    slide_material = "GLASS"
    focus_method = FocusMethod.AUTO
    slide_size_without_label = SizeMm(25, 50)
    image_coordinate_system_rotation: float = 180


defaults = Defaults()



from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Tuple
from wsidicom.config import settings
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.instance.instance import WsiInstance
from wsidicom.series.pyramid import Pyramid


class Pyramids:
    """A collection of pyramids."""

    def __init__(self, pyramids: Iterable[Pyramid]):
        self._pyramids = list(pyramids)

    def __getitem__(self, index: int) -> Pyramid:
        """Get pyramid by index.

        Parameters
        ----------
        index: int
            Index of pyramid to get

        Returns
        -------
        Pyramid
            The pyramid at index in the pyramids.
        """
        return self._pyramids[index]

    def __len__(self) -> int:
        return len(self._pyramids)

    def __iter__(self):
        return iter(self._pyramids)

    def get(self, index: int) -> Pyramid:
        """Get pyramid by index.

        Parameters
        ----------
        index: int
            Index of pyramid to get

        Returns
        -------
        Pyramid
            The pyramid at index in the pyramids.
        """
        try:
            return self[index]
        except IndexError:
            raise WsiDicomNotFoundError(f"Pyramid index {index}", "pyramids")

    @classmethod
    def open(cls, instances: Iterable[WsiInstance]) -> "Pyramids":
        """Return Pyramids object created from wsi instances.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            Instances to create pyramids from.

        Returns
        -------
        Pyramids
            Pyramids created from wsi instances. Each pyramid has the same
            image coordinate system and depth of field.
        """
        instances_grouped_by_pyramid = cls._group_instances_into_pyramids(instances)
        return cls(
            [Pyramid.open(instances) for instances in instances_grouped_by_pyramid]
        )

    @classmethod
    def _group_instances_into_pyramids(
        cls, instances: Iterable[WsiInstance]
    ) -> Iterable[List[WsiInstance]]:
        """Return instances grouped and sorted by pyramid.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            Instances to group by pyramid

        Returns
        -------
        Iterable[List[WsiInstance]]:
            Instances grouped by pyramid.

        """
        grouped_instances: Dict[
            Tuple[Optional[Tuple[float, float, float]], Optional[Tuple[int, float]]],
            List[WsiInstance],
        ] = defaultdict(list)
        for instance in sorted(
            instances, key=lambda x: x.size.to_tuple(), reverse=True
        ):
            if instance.image_coordinate_system is not None:
                image_coordinate_system = (
                    instance.image_coordinate_system.origin.x,
                    instance.image_coordinate_system.origin.y,
                    instance.image_coordinate_system.rotation,
                )
            else:
                image_coordinate_system = None
            if instance.ext_depth_of_field:
                assert instance.ext_depth_of_field_planes is not None
                assert instance.ext_depth_of_field_plane_distance is not None
                ext_depth_of_field = (
                    instance.ext_depth_of_field_planes,
                    instance.ext_depth_of_field_plane_distance,
                )
            else:
                ext_depth_of_field = None

            if instance.image_coordinate_system is not None:
                existing_group_match = next(
                    (
                        (image_cs, ext_dof)
                        for (image_cs, ext_dof), group in grouped_instances.items()
                        if (
                            all(
                                instance.image_coordinate_system.origin_and_rotation_match(
                                    inst.image_coordinate_system,
                                    origin_threshold=settings.pyramids_origin_threshold,
                                )
                                for inst in group
                            )
                            and ext_depth_of_field == ext_dof
                        )
                    ),
                    None,
                )
                if existing_group_match is not None:
                    image_coordinate_system, ext_depth_of_field = existing_group_match

            grouped_instances[image_coordinate_system, ext_depth_of_field].append(
                instance
            )
        return grouped_instances.values()



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Slide model."""

import logging
from typing import Any, Dict, Type

from marshmallow import fields, post_load, pre_dump
from pydicom.valuerep import VR

from wsidicom.conceptcode import ContainerTypeCode
from wsidicom.metadata.sample import SlideSample, SpecimenIdentifier
from wsidicom.metadata.schema.dicom.defaults import defaults
from wsidicom.metadata.schema.dicom.fields import (
    DefaultingDicomField,
    IssuerOfIdentifierDicomField,
    SingleCodeSequenceField,
    StringDicomField,
)
from wsidicom.metadata.schema.dicom.sample import (
    SpecimenDescriptionDicomSchema,
)
from wsidicom.metadata.schema.dicom.sample.formatter import SpecimenDicomFormatter
from wsidicom.metadata.schema.dicom.sample.parser import SpecimenDicomParser
from wsidicom.metadata.schema.dicom.schema import ModuleDicomSchema
from wsidicom.metadata.slide import Slide


class SlideDicomSchema(ModuleDicomSchema[Slide]):
    identifier = DefaultingDicomField(
        StringDicomField(value_representation=VR.LO),
        dump_default=defaults.string,
        data_key="ContainerIdentifier",
        allow_none=True,
    )
    issuer_of_identifier = IssuerOfIdentifierDicomField(
        dump_default=[],
        data_key="IssuerOfTheContainerIdentifierSequence",
        allow_none=True,
    )
    samples = fields.List(
        fields.Nested(SpecimenDescriptionDicomSchema()),
        data_key="SpecimenDescriptionSequence",
        allow_none=True,
    )
    container_type = SingleCodeSequenceField(
        ContainerTypeCode,
        data_key="ContainerTypeCodeSequence",
        dump_only=True,
        dump_default=defaults.slide_container_type,
    )

    @property
    def load_type(self) -> Type[Slide]:
        return Slide

    @pre_dump
    def pre_dump(self, slide: Slide, **kwargs):
        # Move staining to samples so that sample field can serialize both
        if slide.samples is None:
            samples = [SlideSample(identifier=defaults.string)]
        else:
            samples = slide.samples
        dicom_samples = [
            SpecimenDicomFormatter.to_dicom(slide_sample, slide.stainings)
            for slide_sample in samples
        ]
        if isinstance(slide.identifier, SpecimenIdentifier):
            identifier = slide.identifier.value
            issuer_of_identifier = slide.identifier.issuer
        else:
            identifier = slide.identifier
            issuer_of_identifier = None
        return {
            "identifier": identifier,
            "issuer_of_identifier": issuer_of_identifier,
            "samples": dicom_samples,
        }

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        dicom_samples = data.pop("samples", None)
        if dicom_samples is not None:
            try:
                samples, stainings = SpecimenDicomParser().parse_descriptions(
                    dicom_samples
                )

            except (AttributeError, ValueError):
                logging.warning(
                    "Failed to parse SpecimenDescriptionSequence", exc_info=True
                )
                samples = None
                stainings = None
        else:
            samples = None
            stainings = None
        data["samples"] = samples
        data["stainings"] = stainings
        issuer_of_identifier = data.pop("issuer_of_identifier", None)
        if issuer_of_identifier is not None:
            data["identifier"] = SpecimenIdentifier(
                value=data["identifier"], issuer=issuer_of_identifier
            )
        return super().post_load(data, **kwargs)

    @property
    def module_name(self) -> str:
        return "slide"



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with base IO class for handling DICOM WSI files."""

import struct
from datetime import datetime
from functools import cached_property
from pathlib import Path
from struct import pack
from typing import Any, BinaryIO, Callable, Dict, Optional, Union, cast

from fsspec.spec import AbstractBufferedFile
from pydicom.dataelem import RawDataElement, convert_raw_data_element
from pydicom.dataset import Dataset, FileMetaDataset, validate_file_meta
from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomIO
from pydicom.filereader import (
    _read_file_meta_info,
    data_element_generator,
    read_partial,
    read_preamble,
)
from pydicom.filewriter import write_dataset, write_file_meta_info, writers
from pydicom.tag import BaseTag, SequenceDelimiterTag, Tag
from pydicom.uid import UID
from pydicom.valuerep import VR
from upath import UPath

from wsidicom.errors import WsiDicomFileError
from wsidicom.tags import ExtendedOffsetTableTag


class WsiDicomIO:
    """Class for reading or writing DICOM WSI to stream."""

    def __init__(
        self,
        stream: Union[BinaryIO, AbstractBufferedFile],
        transfer_syntax: Optional[UID] = None,
        filepath: Optional[Union[str, Path, UPath]] = None,
        owned: bool = False,
    ):
        """
        Create a WsiDicomIO.

        Parameters
        ----------
        stream: BinaryIO
            Stream to use.
        little_endian: bool = True
            If to set the stream to little endian.
        implicit_vr: bool = False
            If to set the stream to implicit VR.
        filepath: Optional[Union[str, Path, UPath]] = None,
            Optional filepath of stream.
        owned: bool = False
            If the stream should be closed by this instance.
        """
        self._stream = cast(BinaryIO, stream)
        self._stream.seek(0)
        self._filepath = UPath(filepath) if filepath else None
        self._owned = owned
        self._dicom_io = DicomIO(self._stream)
        if transfer_syntax is None:
            transfer_syntax = UID(self.file_meta_info.TransferSyntaxUID)
        self._dicom_io.is_little_endian = transfer_syntax.is_little_endian
        self._dicom_io.is_implicit_VR = transfer_syntax.is_implicit_VR
        self.__enter__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __str__(self) -> str:
        return f"{type(self).__name__}({self._filepath})"

    @property
    def owned(self) -> bool:
        """Return True if the stream is owned by this instance."""
        return self._owned

    @property
    def closed(self) -> bool:
        """Return True if the stream is closed."""
        return self._stream.closed

    @property
    def filepath(self) -> Optional[UPath]:
        """Return filepath, if opened from file."""
        return self._filepath

    @property
    def write(self) -> Callable[[bytes], int]:
        return self._stream.write

    @property
    def seek(self):
        return self._stream.seek

    @property
    def tell(self):
        return self._stream.tell

    @property
    def parent_read(self):
        return self._stream.read

    @property
    def is_little_endian(self):
        return self._dicom_io.is_little_endian

    @property
    def is_implicit_VR(self):
        return self._dicom_io.is_implicit_VR

    @property
    def stream(self):
        return self._stream

    @property
    def is_dicom(self):
        rewind = self.tell()
        self.seek(0)
        self.stream.read(128)  # preamble
        is_dicom = self.stream.read(4) == b"DICM"
        self.seek(rewind)
        return is_dicom

    @property
    def media_storage_sop_class_uid(self) -> UID:
        """Read Media Storage SOP Class UID from file meta info."""
        return self.file_meta_info.MediaStorageSOPClassUID

    @cached_property
    def file_meta_info(self) -> FileMetaDataset:
        """Read file meta info from stream."""
        self.seek(0)
        try:
            read_preamble(self._stream, False)
            file_meta_info = _read_file_meta_info(self._stream)
        except InvalidDicomError:
            raise WsiDicomFileError(str(self), "is not a DICOM file or stream.")
        finally:
            self.seek(0)
        return file_meta_info

    def read_dataset(self, force: bool = False) -> Dataset:
        """Read dataset, excluding EOT and PixelData from stream."""
        extended_offset_table_tag = ExtendedOffsetTableTag

        def _stop_at(tag: BaseTag, vr: Optional[str], length: int) -> bool:
            return tag >= extended_offset_table_tag

        self.seek(0)
        return read_partial(
            self._stream,
            _stop_at,
            defer_size=None,
            force=force,
            specific_tags=None,
        )

    def read_tag(self) -> BaseTag:
        """Read tag from stream."""
        return Tag(self._dicom_io.read_tag())

    def read_tag_length(self, long: bool) -> int:
        """Read tag length."""
        if not long and not self._dicom_io.is_implicit_VR:
            return self._dicom_io.read_US()
        return self._dicom_io.read_UL()

    def read_tag_vr(self) -> Optional[bytes]:
        """Read tag VR if implicit VR."""
        if not self._dicom_io.is_implicit_VR:
            vr = self.stream.read(4)
            return vr[0:2]

    def read_UL(self) -> int:
        """Read unsigned long integer (32 bits)."""
        return self._dicom_io.read_UL()

    def read(self, size: int, need_exact_length: bool = False) -> bytes:
        """Read bytes from stream."""
        data = self._stream.read(size)
        if need_exact_length and len(data) != size:
            raise EOFError()
        return data

    def check_tag_and_length(
        self, tag: BaseTag, length: int, with_vr: bool, long: bool
    ) -> None:
        """Check if tag at position is expected tag with expected length.

        Parameters
        ----------
        tag: BaseTag
            Expected tag.
        length: int
            Expected length.
        with_vr: bool
            If tag is expected to have VR.
        long: bool
            If length is expected to be long.

        """
        try:
            read_tag = self._dicom_io.read_tag()
            if tag != read_tag:
                raise WsiDicomFileError(
                    str(self), f"Found tag {read_tag} expected {tag}."
                )
            if with_vr:
                if self._dicom_io.is_implicit_VR:
                    raise WsiDicomFileError(str(self), "Expected VR, but implicit VR.")
                self.read_tag_vr()
            read_length = self.read_tag_length(long)
            if length != read_length:
                raise WsiDicomFileError(
                    str(self), f"Found length {read_length} expected {length}."
                )
        except struct.error:
            raise WsiDicomFileError(str(self), "Failed to unpack data.")

    def read_sequence_delimiter(self):
        """Check if last read tag was a sequence delimiter.
        Raises WsiDicomFileError otherwise.
        """
        TAG_BYTES = 4
        self.seek(-TAG_BYTES, 1)
        if self._dicom_io.read_tag() != SequenceDelimiterTag:
            raise WsiDicomFileError(str(self), "No sequence delimiter tag")

    def write_unsigned_long_long(self, value: int):
        """Write unsigned long long integer (64 bits).

        Parameters
        ----------
        value: int
            Value to write.

        """
        if self._dicom_io.is_little_endian:
            format = "<Q"
        else:
            format = ">Q"
        self.write(pack(format, value))

    def write_tag(self, tag: BaseTag):
        """Write tag to stream.

        Parameters
        ----------
        tag: BaseTag
            Tag to write.

        """
        self._dicom_io.write_tag(tag)

    def write_UL(self, value: int):
        """Write unsigned long integer (32 bits).

        Parameters
        ----------
        value: int
            Value to write.

        """
        self._dicom_io.write_UL(value)

    def write_tag_of_vr_and_length(
        self, tag: BaseTag, value_representation: str, length: Optional[int] = None
    ):
        """Write tag, tag VR and length.

        Parameters
        ----------
        tag: str
            Name of tag to write.
        value_representation: str.
            Value representation (VR) of tag to write.
        length: Optional[int] = None
            Length of data after tag. 'Unspecified' (0xFFFFFFFF) if None.

        """
        self._dicom_io.write_tag(Tag(tag))
        if not self._dicom_io.is_implicit_VR:
            self.write(bytes(value_representation, "iso8859"))
            self._dicom_io.write_US(0)
        if length is not None:
            self._dicom_io.write_UL(length)
        else:
            self._dicom_io.write_UL(0xFFFFFFFF)

    def write_preamble(self):
        """Write DICOM preamble."""
        self.seek(0)
        preamble = b"\x00" * 128
        self.write(preamble)
        self.write(b"DICM")

    def write_file_meta_info(
        self, instance_uid: UID, sop_class_uid: UID, transfer_syntax: UID
    ):
        """Write file meta info.

        Parameters
        ----------
        instance_uid: UID
            SOP Instance UID.
        sop_class_uid: UID
            SOP Class UID.

        """
        meta = FileMetaDataset()
        meta.TransferSyntaxUID = transfer_syntax
        meta.MediaStorageSOPInstanceUID = instance_uid
        meta.MediaStorageSOPClassUID = sop_class_uid
        validate_file_meta(meta)
        write_file_meta_info(self._dicom_io, meta)

    def write_dataset(self, dataset: Dataset, content_datetime: datetime):
        """Write dataset to stream.

        Parameters
        ----------
        dataset: Dataset
            Dataset to write.

        """
        dataset.ContentDate = datetime.date(content_datetime).strftime("%Y%m%d")
        dataset.ContentTime = datetime.time(content_datetime).strftime("%H%M%S.%f")
        write_dataset(self._dicom_io, dataset)

    def close(self, force: Optional[bool] = False) -> None:
        """Close stream if owned by instance or forced."""
        if self._owned or force:
            self._stream.close()

    def update_dataset(self, dataset_start: int, update: Dict[BaseTag, Any]):
        """Update dataset in place.

        The element value representation must allow padding,
        and values to replace should be padded to the same length or longer than the
        replacement value.


        Parameters
        ----------
        dataset_start: int
            Position of dataset start.
        update: Dict[BaseTag, Any]
            Dictionary with tags to update and their new values.
        """

        rewind = self.tell()
        self.seek(dataset_start)
        for element in data_element_generator(
            self._stream,
            self.is_implicit_VR,
            self.is_little_endian,
            specific_tags=list(update.keys()),
        ):
            if element.tag not in update.keys():
                # generator can include `Specific Character Set` element
                continue
            if not isinstance(element, RawDataElement):
                raise ValueError("Can only update raw data elements.")
            element_value_position = element.value_tell
            length = element.length
            if isinstance(element, RawDataElement):
                element = convert_raw_data_element(element)
            element.value = update[BaseTag(element.tag)]
            self.seek(element_value_position)
            writer, param = writers[VR(element.VR)]
            if param is None:
                param = []
            writer(self, element, *param)
            if self.tell() > element_value_position + length:
                raise ValueError("Updated element is longer than original.")
        self.seek(rewind)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from wsidicom.file.io.frame_index.basic import (
    BasicOffsetTableFrameIndexParser,
    EmptyBasicTableOffsetException,
)
from wsidicom.file.io.frame_index.extended import ExtendedOffsetFrameIndexParser
from wsidicom.file.io.frame_index.native_pixel_data import (
    NativePixelDataFrameIndexParser,
)
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType
from wsidicom.file.io.frame_index.offset_table_writer import (
    BotWriter,
    EotWriter,
    OffsetTableWriter,
)
from wsidicom.file.io.frame_index.parser import FrameIndexParser
from wsidicom.file.io.frame_index.pixel_data import PixelDataFrameIndexParser

__all__ = [
    "BasicOffsetTableFrameIndexParser",
    "PixelDataFrameIndexParser",
    "ExtendedOffsetFrameIndexParser",
    "FrameIndexParser",
    "NativePixelDataFrameIndexParser",
    "OffsetTableType",
    "BotWriter",
    "EotWriter",
    "OffsetTableWriter",
    "EmptyBasicTableOffsetException",
]



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Frame index for BOT, parsing the positions from the BOT."""

from typing import List, Optional, Tuple

from wsidicom.file.io.frame_index.offset_table import OffsetTableFrameIndexParser
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType


class EmptyBasicTableOffsetException(Exception):
    """Exception raised when BOT was empty."""

    pass


class BasicOffsetTableFrameIndexParser(OffsetTableFrameIndexParser):
    @property
    def offset_table_type(self) -> OffsetTableType:
        return OffsetTableType.BASIC

    @property
    def bytes_per_item(self) -> int:
        return 4

    @property
    def mode(self) -> str:
        return "<L"

    def _get_index(self) -> List[Tuple[int, int]]:
        """Get frame positions and length from bot."""
        self._validate_pixel_data_start()
        table = self._read_table()
        pixels_start = self._file.tell()
        assert table is not None
        return self._parse_table(table, pixels_start)

    def _get_pixels_start(self) -> int:
        self._validate_pixel_data_start()
        bot_length = self._read_bot_length()
        if bot_length is None:
            raise EmptyBasicTableOffsetException()
        return self._file.tell()

    def _read_table(self) -> Optional[bytes]:
        """Read basic table offset (BOT). Returns None if BOT is empty.

        Returns
        -------
        Optional[bytes]
            BOT in bytes.
        """
        bot_length = self._read_bot_length()
        if bot_length is None:
            raise EmptyBasicTableOffsetException()
        return self._file.read(bot_length, need_exact_length=True)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from collections.abc import Callable, Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Optional, TypeVar

ReturnType = TypeVar("ReturnType")


class ConditionalThreadPoolExecutor(ThreadPoolExecutor):
    """ThreadPoolExecutor that uses a single thread if max workers is 1."""

    def __init__(
        self, max_workers: Optional[int] = None, force_iteration: bool = False, **kwargs
    ) -> None:
        self._force_iteration = force_iteration
        super().__init__(max_workers, **kwargs)

    def map(
        self,
        fn: Callable[..., ReturnType],
        *iterables: Iterable[Any],
        timeout: Optional[float] = None,
        chunksize: int = 1
    ) -> Iterator[ReturnType]:
        if self._max_workers == 1:
            if self._force_iteration:
                # Make sure items are iterated through
                return iter(list(map(fn, *iterables)))
            return map(fn, *iterables)
        return super().map(fn, *iterables, timeout=timeout, chunksize=chunksize)
